/*
 * $Id: AmRtpStream.cpp,v 1.54 2004/06/16 12:30:46 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmRtpStream.h"
#include "AmRtpPacket.h"
#include "AmRtpScheduler.h"
#include "AmIcmpWatcher.h"
#include "AmConfig.h"
#include "AmPlugIn.h"
#include "AmAudio.h"
#include "AmUtils.h"
#include "log.h"

#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>       
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <set>
using std::set;




void RtpSync::begin(unsigned int ts, AmRtpScheduler* sched)
{
    if(!init){
	last_time = begin_time = sched->getCurrentTime();
	last_ts = begin_ts = ts;
    }
    else {
	unsigned int ts_diff = ts - last_ts;
	last_ts = ts;
	
	struct timeval time_diff;
	time_diff.tv_sec = ts_diff/8000;
	time_diff.tv_usec = (ts_diff - time_diff.tv_sec*8000)*125;
	timeradd(&time_diff,&last_time,&last_time);
    }
}

void RtpSync::sync(AmRtpScheduler* sched)
{
    if(!init){
	init = true;
    }
    else {
	AmCondition<bool> fired(false);
	AmRtpTimer timer(&fired);
	timer.time = last_time;
	sched->addTimer(timer);
	fired.wait_for();
    }
}

unsigned int RtpSync::computeTimestamp(AmRtpScheduler* sched)
{
    struct timeval tv;
    if(!init){
	gettimeofday(&tv,NULL);
	// pseudo random value
	return rand() + tv.tv_sec + tv.tv_usec;
    }
    else {
	tv = sched->getCurrentTime();
	timersub(&tv,&begin_time,&tv);
	return (begin_ts + tv.tv_sec*8000 + tv.tv_usec/125);
    }
}


/*
 * This function must be called before setLocalPort, because
 * setLocalPort will bind the socket and it will be not
 * possible to change the IP later
 */
void AmRtpStream::setLocalIP(const string& ip)
{
#ifdef SUPPORT_IPV6
    if (!inet_aton_v6(ip.c_str(), &l_saddr)) {
#else
    if (!inet_aton(ip.c_str(), (in_addr*)&l_saddr.sin_addr.s_addr)) {
#endif
	    throw string ("Invalid IP address.");
    }
}

void AmRtpStream::setLocalPort()
{
    if(l_port)
 	return;
    
    int sd=0;
#ifdef SUPPORT_IPV6
    if((sd = socket(l_saddr.ss_family,SOCK_DGRAM,0)) == -1){
#else
    if((sd = socket(PF_INET,SOCK_DGRAM,0)) == -1){
#endif
	ERROR("%s\n",strerror(errno));
	throw string ("while creating new socket.");
    }

    int true_opt = 1;
    if(ioctl(sd, FIONBIO , &true_opt) == -1){
	ERROR("%s\n",strerror(errno));
	close(sd);
	throw string ("while setting socket non blocking.");
    }

    int retry = 10;
    unsigned short port = 0;
    DBG("setting random local port\n");
    for(;retry; --retry){

	port = (rand() % (AmConfig::RtpHighPort - AmConfig::RtpLowPort + 1)) +  AmConfig::RtpLowPort;

	// only even port number
	port &= 0xfffe;

#ifdef SUPPORT_IPV6
	set_port_v6(&l_saddr,port);
	if(!bind(sd,(const struct sockaddr*)&l_saddr,
		 sizeof(struct sockaddr_storage)))
	    break;
#else	
	l_saddr.sin_port = htons(port);
	if(!bind(sd,(const struct sockaddr*)&l_saddr,
		 sizeof(struct sockaddr_in)))
	    break;
#endif
    }
    if (!retry){
	ERROR("could not find a random local port !\n");
	close(sd);
	throw string("while setting random local port");
    }

    if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
		  (void*)&true_opt, sizeof (true_opt)) == -1) {

	ERROR("%s\n",strerror(errno));
	close(sd);
	throw string ("while setting local address reusable.");
    }
    
    l_sd = sd;
    l_port = port;
    AmIcmpWatcher::instance()->addStream(l_port,this);
    DBG("local rtp port set to %i\n",l_port);
}

int AmRtpStream::send( unsigned int ts, AmAudio* audio_provider, int samples )
{
    char buffer[1024];

    int i = audio_provider->get(ts,(unsigned char*)buffer,samples);
    if(i > 0){

	if(mute)
	    return 0;

	AmRtpPacket rp;
	rp.payload = payload;
	rp.marker = false;
	rp.sequence = sequence++;
	rp.timestamp = ts;   
	rp.ssrc = l_ssrc;

	rp.compile((unsigned char*)buffer,i);

	send_s.begin(ts,sched);
	if(rp.send(l_sd,&r_saddr) < 0){
	    ERROR("while sending RTP packet.");
	    return -1;
	}
    }
    
    return i;
}


int AmRtpStream::receive( unsigned int ts, AmAudio* audio_provider, bool record,
			  AmRtpPacketTracer* tracer )
{

    int recv_size = 0;
#ifdef SUPPORT_IPV6
    struct sockaddr_storage recv_addr;
#else
    struct sockaddr_in recv_addr;
#endif

    recv_s.begin(ts,sched);

    struct timeval now;
    struct timeval diff;
    gettimeofday(&now,NULL);
    timersub(&now,&last_recv_time,&diff);

    if(diff.tv_sec > DEAD_RTP_TIME){
	WARN("Last received packet is too old: ending session.\n");
	thread->stop();
	return -1;
    }

    if(blocking)
	recv_s.sync(sched);

    if(mute)
	return 0;

    while(true){

	AmRtpPacket rp;

	if(rp.recv(l_sd,&recv_addr) <= 0)
	    break;

	if(tracer)
	    tracer->write_packet(ts,&recv_addr,&rp);

	// symmetric RTP
#ifndef SUPPORT_IPV6
 	if( passive && ((recv_addr.sin_port != r_saddr.sin_port)
			|| (recv_addr.sin_addr.s_addr 
			    != r_saddr.sin_addr.s_addr)) ) {
	    
	    string addr_str = get_addr_str(recv_addr.sin_addr);
	    int port = ntohs(recv_addr.sin_port);
	    setRAddr(addr_str,port);
	    DBG("Symmetric RTP: setting new remote address: %s:%i\n",
		addr_str.c_str(),port);
 	}
#endif
	if(!record)
	    continue;

	if(rp.parse() == -1){
	    ERROR("while parsing RTP packet.\n");
	    return -1;
	}

	assert(audio_provider);
	audio_provider->begin_talk = ((last_payload == 13) 
				      || rp.marker);
	last_payload = payload;

	if((payload != rp.payload)
	   || !rp.getDataSize()){
	    continue;
	}

	if(!recv_offset_i){
	    recv_offset = ts - rp.timestamp;
	    recv_offset_i = true;
	}

	assert(rp.getData());
	if(audio_provider->put( recv_offset + rp.timestamp,
				rp.getData(),rp.getDataSize() ) < 0){
	    ERROR("while transmitting RTP datas to audio engine.\n");
	    return -1;
	}
	
	last_recv_time = now;
	recv_size += rp.getDataSize();
    }

    return recv_size;
}

unsigned int AmRtpStream::computeTimestamp()
{
    return recv_s.computeTimestamp(sched);
}

AmRtpStream::AmRtpStream(AmThread* _t) 
    : runcond(0), 
      r_port(0),
      l_port(0),
      l_sd(0), 
      recv_offset_i(false), 
      r_ssrc_i(false),
      user_ts(0), 
      blocking(true), 
      thread(_t),
      passive(false),
      first_recved(false),
      mute(false)
{
    assert(thread);
#ifdef SUPPORT_IPV6
    memset(&r_saddr,0,sizeof(struct sockaddr_storage));
    memset(&l_saddr,0,sizeof(struct sockaddr_storage));
#else
    memset(&r_saddr,0,sizeof(struct sockaddr_in));
    memset(&l_saddr,0,sizeof(struct sockaddr_in));
#endif
    sched = AmRtpScheduler::instance();

#ifndef SUPPORT_IPV6
    /* By default we listen on all interfaces */
    l_saddr.sin_family = AF_INET;
    l_saddr.sin_addr.s_addr = INADDR_ANY;
#endif
}

AmRtpStream::~AmRtpStream()
{
    if(l_sd){
	AmIcmpWatcher::instance()->removeStream(l_port);
	close(l_sd);
    }
}

int AmRtpStream::getLocalPort()
{
    if(!l_port)
	setLocalPort();

    return l_port;
}

int AmRtpStream::getRPort()
{
    return r_port;
}

void AmRtpStream::setRAddr(const string& addr, unsigned short port)
{
    DBG("RTP remote address set to %s:%u\n",addr.c_str(),port);

#ifdef SUPPORT_IPV6
    struct sockaddr_storage ss;
    if(!inet_aton_v6(addr.c_str(),&ss)){
 	ERROR("address not valid (host: %s)\n",addr.c_str());
 	throw string("invalid address");
    }
    
    memcpy(&r_saddr,&ss,sizeof(struct sockaddr_storage));
    set_port_v6(&r_saddr,port);
#else
    struct sockaddr_in sa;
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    
    if(!inet_aton(addr.c_str(),&sa.sin_addr)){
 	ERROR("address not valid (host: %s)\n",addr.c_str());
 	throw string("invalid address");
    }

    memcpy(&r_saddr,&sa,sizeof(struct sockaddr_in));
#endif

    r_host = addr;
    r_port = port;

#ifndef SUPPORT_IPV6
    mute = (sa.sin_addr.s_addr == 0);
#endif
}

void AmRtpStream::start(const SdpPayload& sdp_payload)
{
    payload = sdp_payload.payload_type;
    int_payload = sdp_payload.int_pt;
    last_payload = payload;
    format_parameters = sdp_payload.sdp_format_parameters;
    resume();
}

void AmRtpStream::pause()
{
    runcond.set(false);
}

void AmRtpStream::resume()
{
    user_ts = computeTimestamp();
    gettimeofday(&last_recv_time,NULL);
    runcond.set(true);
}

void AmRtpStream::icmpError()
{
    if(!passive || first_recved.get()){
	AmIcmpWatcher::instance()->removeStream(l_port);
	thread->stop();
    }
}

void AmRtpStream::play(AmAudio* audio_provider, int max_play_time)
{
    //TODO: do not ignore max_play_time

    if(!audio_provider){
	ERROR("AmRtpStream::play: audio_provider is NULL.\n");
	return;
    }

    if(!runcond.get())
	return;

    blocking = true;
    audio_provider->out.reset(new AmAudioRtpFormat(int_payload, format_parameters));

    /* number of samples in a time frame */
    int inc_samples = audio_provider->out->frame_size; //(audio_provider->out->rate * audio_provider->out->frame_length) / 1000;
    
    while( true )
    {
	// send
	if( !runcond.get() || (send(user_ts,audio_provider,inc_samples) < 0) )
	    break;
	
	// flush receive buffers
    	receive(user_ts,audio_provider,false);

	user_ts += inc_samples;
	thread->onIdle();
    }
    audio_provider->close();
}

int AmRtpStream::record(AmAudio* audio_provider, int max_rec_time)
{
    int  rec_time;

    if(!audio_provider){
	ERROR("AmRtpStream::record: audio_provider is NULL.\n");
	return -1;
    }

    if(!runcond.get())
	return 0;

    blocking = true;
    audio_provider->in.reset(new AmAudioRtpFormat(int_payload, format_parameters));

    if(audio_provider->out->rate < 0) 
	audio_provider->out->rate = audio_provider->in->rate;
    if(audio_provider->out->channels < 0) 
	audio_provider->out->channels = audio_provider->in->channels;

    /* number of samples in a frame */
    int inc_samples = audio_provider->in->frame_size; //(audio_provider->in->rate * audio_provider->in->frame_length) / 1000;
        
    rec_time = audio_provider->in->rate * max_rec_time;

    while(runcond.get() && (rec_time > 0)){
	
	if(receive(user_ts,audio_provider,true) < 0)
	    break;

	user_ts += inc_samples;
	rec_time -= inc_samples;
	thread->onIdle();
    }

    audio_provider->close();
    return audio_provider->getDataSize();
}

int  AmRtpStream::duplex( AmAudio* audio_play, AmAudio* audio_rec, int max_time )
{
    int rec_time;

    if(!audio_play || !audio_rec){
	ERROR("AmRtpStream::duplex: audio_play or audio_rec is NULL.\n");
	return -1;
    }

    if(!runcond.get())
	return 0;

    blocking = true;
    audio_rec->in.reset(new AmAudioRtpFormat(int_payload, format_parameters));
    audio_play->out.reset(new AmAudioRtpFormat(int_payload, format_parameters));

    if(audio_rec->out->rate < 0)
	audio_rec->out->rate = audio_rec->in->rate;
    if(audio_rec->out->channels < 0)
	audio_rec->out->channels = audio_rec->in->channels;

    /* number of samples in a frame */
    int inc_samples = audio_rec->in->frame_size;  //(audio_rec->in->rate * FRAME_LENGTH_MS) / 1000;
    rec_time = audio_rec->in->rate * max_time;

    /* TODO: support different frame size for in & out codecs */
    while( ((max_time < 0) || (rec_time > 0)) && runcond.get() ){

 	if(send(user_ts,audio_play,inc_samples) < 0)
	    break;

	if(receive(user_ts,audio_rec,true) < 0)
	    break;

	user_ts += inc_samples;
	rec_time -= inc_samples;
	thread->onIdle();
    }

    audio_play->close();
    audio_rec->close();

    return audio_rec->getDataSize();
}

void AmRtpStream::play(const string& audio_file, int max_play_time)
{
    AmAudioFile wav_file("Wav");
    if(wav_file.open(audio_file.c_str(),AmAudioFile::Read)){
 	ERROR("AmRtpStream::play: Cannot open file\n");
 	return;
    }
    
    play(&wav_file,max_play_time);
}


int AmRtpStream::record(const string& audio_file, int max_rec_time)
{
    unsigned int dot_pos = audio_file.rfind('.');
    string ext = audio_file.substr(dot_pos+1);

    if(ext.empty()){
	ERROR("AmRtpStream::record: could not extract file extension "
	      "(file: %s)\n",audio_file.c_str());
	return -1;
    }

    AmPlugIn* plug_in = AmPlugIn::instance();
    amci_inoutfmt_t* fmt = plug_in->fileFormat("",ext);

    if(!fmt){
	ERROR("AmRtpStream::record: plug-in for .%s files not found\n",
	      ext.c_str());
	return -1;
    }

    AmAudioFile af(fmt->name);// uses default subtype
    if(af.open(audio_file.c_str(),AmAudioFile::Write)){
	ERROR("AmRtpStream::record(): couldn't open %s for writing\n",
	      audio_file.c_str());
	return -1;
    }
    
    return record(&af,max_rec_time);
}

AmRtpStreamSet::AmRtpStreamSet()
    : sched(0),ts_set(0),nb_packet(0)
{
    sched = AmRtpScheduler::instance();
    gettimeofday(&ts_clock_begin,NULL);
}

AmRtpStreamSet::~AmRtpStreamSet()
{
}

void AmRtpStreamSet::addStream(AmRtpStream* str, const AmRtpStreamInfo& info)
{
    if(!str->runcond.get()){
	ERROR("stream has not been started: it won't be added to the set !\n");
	return;
    }

    std::pair<StreamSet::iterator, bool> res = 
	stream_set.insert(std::make_pair(str,info));

    if(res.second){

	StreamSet::iterator elmt = res.first;

	if(elmt->second.audio_play)
	    elmt->second.audio_play->out.reset(new AmAudioRtpFormat(str->int_payload, str->format_parameters));

	if(elmt->second.audio_rec)
	    elmt->second.audio_rec->in.reset(new AmAudioRtpFormat(str->int_payload, str->format_parameters));
    }
}

void AmRtpStreamSet::removeStream(AmRtpStream* str)
{
    StreamSet::iterator it = stream_set.find(str);

    if(it != stream_set.end()){

 	delete it->second.audio_play;
 	if(it->second.audio_play != it->second.audio_rec)
 	    delete it->second.audio_rec;
	stream_set.erase(it);
    }
}

void AmRtpStreamSet::setStrInfo(AmRtpStream* str, const AmRtpStreamInfo& info)
{
    StreamSet::iterator it = stream_set.find(str);

    if(it != stream_set.end()){

 	delete it->second.audio_play;
 	if(it->second.audio_play != it->second.audio_rec)
 	    delete it->second.audio_rec;
	
	it->second.type       = info.type;
	it->second.audio_play = info.audio_play;
	it->second.audio_rec  = info.audio_rec;

	if(it->second.audio_play)
	    it->second.audio_play->out.reset(new AmAudioRtpFormat(str->int_payload, str->format_parameters));

	if(it->second.audio_rec)
	    it->second.audio_rec->in.reset(new AmAudioRtpFormat(str->int_payload, str->format_parameters));
    }
}

int AmRtpStreamSet::exchange()
{
     if(stream_set.empty()){
 	ERROR("exchange has been called with an empty set !\n");
 	return -1;
     }

     StreamSet::iterator it = stream_set.begin();
     int ret = 0;
     try {
	 if(nb_packet){
	     struct timeval ts_clock;
	     gettimeofday(&ts_clock,NULL);
	     timersub(&ts_clock,&ts_clock_begin,&ts_clock);
	     double ts_ms = ts_clock.tv_sec*1000 + ts_clock.tv_usec/1000;

	     double m_inter = (nb_packet ? ts_ms/double(nb_packet) : 0.0);
	     int diff = int(double(nb_packet)*20.0 - ts_ms);
	 
	     if(!(nb_packet & 127)){
		 DBG("diff = %i; interval = %u; ts_ms = %u; # packets: %u\n",
		     diff,(unsigned int)m_inter,(unsigned int)ts_ms,nb_packet);
	     }
	 }

	 for(it = stream_set.begin(); it != stream_set.end(); ++it){
	     
	     AmRtpStream*      rtp_str = it->first;
	     AmRtpStreamInfo&  str_info = it->second;

	     if(!str_info.ts_offset_i){
		 rtp_str->user_ts = rtp_str->computeTimestamp();
		 str_info.ts_offset = rtp_str->user_ts - ts_set;
		 str_info.ts_offset_i = true;
	     }
	  
	     rtp_str->blocking = false;

	     if(str_info.audio_rec){
	         rtp_str->receive(ts_set+str_info.ts_offset,
				  str_info.audio_rec,true,tracer.get());
	     }
	     else {
		 rtp_str->receive(ts_set+str_info.ts_offset,
				  NULL,false);
	     }
	 }

	 for(it = stream_set.begin(); it != stream_set.end(); ++it){
	     
	     AmRtpStream*      rtp_str = it->first;
	     AmRtpStreamInfo&  str_info = it->second;

	     if(str_info.audio_play){
	         if(rtp_str->send(ts_set,
				  str_info.audio_play, 160) < 0) 
		     throw string("while sending...");
	     }
	 }

	 sync.begin(ts_set,sched);
	 sync.sync(sched);

	 ts_set += 160;
	 nb_packet++;
     }
     catch(const string& err){
 	ERROR("%s\n",err.c_str());
 	ret = -1;
     }
     
     return ret;
}

void AmRtpStreamSet::enableRTPTrace(const string& trace_file)
{
    tracer.reset(new AmRtpPacketTracer(trace_file));
}

AmRtpStreamInfo::AmRtpStreamInfo( StreamType type, 
				  AmAudio* audio_play, 
				  AmAudio* audio_rec)
    : type(type), 
      audio_play(audio_play), 
      audio_rec(audio_rec),
      ts_offset_i(false),
      ts_offset(0)
{
}
