/*
 * $Id: AmServer.cpp,v 1.25 2004/03/14 17:10:47 jiri Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmServer.h"
#include "log.h"
#include "AmSession.h"
#include "AmRequest.h"
#include "AmMail.h"
#include "AmUtils.h"
#include "AmPlugIn.h"
#include "AmConfig.h"
#include "AmRtpScheduler.h"
#include "AmIcmpWatcher.h"

#include <errno.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#define READ_PARAMETER(p)\
             READ_FIFO_PARAMETER(fifo_stream,p,line_buf,MAX_LINE_SIZE)

AmServer* AmServer::_instance;

AmServer* AmServer::instance()
{
    if(!_instance)
	_instance = new AmServer();
    return _instance;
}

AmServer::AmServer()
    : fifo_stream(0)
{
}

AmServer::~AmServer()
{
    if(fifo_stream)
	fclose(fifo_stream);
}

int AmServer::get_line(char* str, size_t len)
{
    return fifo_get_line(fifo_stream,str,len);
}

int AmServer::get_lines(char* str, size_t len)
{
    return fifo_get_lines(fifo_stream,str,len);
}


void AmServer::consume_request()
{
    char line_buf[MAX_LINE_SIZE];
    while( get_line(line_buf,MAX_LINE_SIZE) != 0 ) {
	    ERROR("Consumed: %s\n",line_buf);
    }
}

char* AmServer::get_next_tok(char*& str)
{
    char* start=str;
    while(*start==' ') start++;
    
    char* s=start;
    while(*s){
	if(*s == ' '){
	    *s='\0';
	    str=s+1;
	    return start;
	}
	s++;
    }
    str = s;
    return ( s > start ? start : 0 );
}

int is_main=1;
int child_pid=0;

void AmServer::run(const char * fifo_name)
{
    if( (mkfifo(fifo_name,FIFO_PERM)<0) && (errno!=EEXIST) ) {
	ERROR("while creating fifo `%s': %s \n",fifo_name,strerror(errno));
	return;
    }

    // avoid the FIFO reaches EOF...
    if((child_pid=fork())>0){
	FILE* fifo_write=0;
        if(!(fifo_write=fopen(fifo_name, "w"))){
	    ERROR("while opening fifo `%s': %s\n",fifo_name,strerror(errno));
	    kill(child_pid,SIGTERM);
	}
	waitpid(child_pid,0,0);
	if(fifo_write)
	    fclose(fifo_write);
	return;
    }

    is_main = 0;

    if( !(fifo_stream = fopen(fifo_name,"r")) ) {
	ERROR("while opening fifo `%s': %s\n",fifo_name,strerror(errno));
	return;
    }

    DBG("Loading application plug-ins\n");
    if(AmPlugIn::instance()->load(AmConfig::PlugInPath + "/apps", AmPlugIn::App))
	return;

    DBG("Loading audio plug-ins\n");
    if(AmPlugIn::instance()->load(AmConfig::PlugInPath + "/audio", AmPlugIn::Audio))
	return;

    DBG("Starting session container\n");
    AmSessionContainer::instance()->start();

    DBG("Starting mailer\n");
    AmMailDeamon::instance()->start();

    DBG("Starting RTP scheduler\n");
    AmRtpScheduler::instance()->start();


    DBG("Starting ICMP watcher\n");
    AmIcmpWatcher::instance()->start();

    INFO("Sems started\n");

    char line_buf[MAX_LINE_SIZE];
    string version;
    string fct_name;
    string cmd;
    unsigned int pos;

    while(true){

	try {

	    
	    READ_PARAMETER(version);
	    if (version == "") {
		    // some odd trailer -- ignore
		    ERROR("odd trailer\n");
		    continue;
	    }
	    if(version != FIFO_VERSION){
		throw string("wrong FIFO Interface version:"+version);
	    }

	    READ_PARAMETER(fct_name);

	    if( ((pos = fct_name.find('.')) == string::npos)
		// || (pos+1 >= fct_name.length())
		)
		throw string("bad command: '" + fct_name + "'. "
			     "Syntax: fct_name.{plugin_name,'bye'}");

	    cmd = fct_name.substr(pos+1,string::npos);
	    fct_name = fct_name.substr(0,pos);

	    if(fct_name != "sip_request"){
		map<string,AmServerFct*>::iterator fct_it;
		if( (fct_it = fct_map.find(fct_name)) == fct_map.end() )
		    throw string("unknown server function: '" + fct_name + "'");

		fct_it->second->execute(fifo_stream,cmd);
		continue;
	    }
	    else
		execute(fifo_stream,cmd);
	}
	catch(const string& err_msg){
	    ERROR("%s\n",err_msg.c_str());
	    consume_request();
	    continue;
	}
    }
}

int  AmServer::registerFct(const string& name,AmServerFct* fct)
{
    if(fct_map.find(name) != fct_map.end()){
	ERROR("AmServer::registerFct: function '%s'"
	      " has already been registered.\n",name.c_str());
	return -1;
    }
    
    fct_map[name] = fct;
    DBG("AmServer::registerFct: function '%s'"
	" has been registered.\n",name.c_str());

    return 0;
}

int  AmServer::execute(FILE* stream, const string& cmd_str)
{
    char line_buf[MAX_LINE_SIZE];
    char body_buf[MAX_BODY_SIZE];
    char hdrs_buf[MAX_BODY_SIZE];
    int  body_len=0, hdrs_len=0;

    AmCmd cmd;

    if(cmd_str.empty())
	throw string("AmServer::execute: FIFO parameter plug-in name missing.");

    cmd.cmd = cmd_str;
    READ_PARAMETER(cmd.method);
    READ_PARAMETER(cmd.user);
    READ_PARAMETER(cmd.email);
    READ_PARAMETER(cmd.domain);
    READ_PARAMETER(cmd.dstip);    // will be taken for UDP's local peer IP & Contact
    READ_PARAMETER(cmd.port);     // will be taken for Contact
    READ_PARAMETER(cmd.r_uri);    // ??? will be taken for the answer's Contact header field ???
    READ_PARAMETER(cmd.from_uri); // will be taken for subsequent request uri
    READ_PARAMETER(cmd.from);
    READ_PARAMETER(cmd.to);
    READ_PARAMETER(cmd.callid);
    READ_PARAMETER(cmd.from_tag);
    READ_PARAMETER(cmd.to_tag);

    string cseq_str;
    READ_PARAMETER(cseq_str);
    
    if(sscanf(cseq_str.c_str(),"%u", &cmd.cseq) != 1){
	throw string("invalid cseq number (") + cseq_str + string(")\n");
    }
    DBG("cseq= <%s>(%i)\n",cseq_str.c_str(),cmd.cseq);
    
    READ_PARAMETER(cmd.key);
    READ_PARAMETER(cmd.route);
    READ_PARAMETER(cmd.next_hop);
    
    if( (hdrs_len = get_lines(hdrs_buf,MAX_BODY_SIZE)) == -1 ) {
	throw string("len(hdrs) > ") + int2str(MAX_BODY_SIZE) + "\n";
    }
    hdrs_buf[hdrs_len]='\0';
    DBG("hdrs: `%s'\n",hdrs_buf);
    cmd.hdrs = hdrs_buf;
    
    if( (body_len = get_lines(body_buf,MAX_BODY_SIZE)) == -1 ) {
	throw string("len(body) > ") + int2str(MAX_BODY_SIZE) + "\n";
    }
    body_buf[body_len]='\0';
    DBG("body: `%.*s'\n",body_len,body_buf);
    
#define IS_EMPTY(p) if(p.empty()) \
                     throw string("invalid FIFO command: " #p " is empty !!!" );
    
    IS_EMPTY(cmd.cmd);
    IS_EMPTY(cmd.from);
    IS_EMPTY(cmd.to);
    IS_EMPTY(cmd.callid);
    IS_EMPTY(cmd.from_tag);
    
    DBG("everything is OK !\n");
    
    AmRequestUAS req(cmd,body_buf);
    req.execute();
    
    return 0;
}







