/*
 * $Id: AmSession.h,v 1.18 2004/03/08 18:56:47 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _AmSession_h_
#define _AmSession_h_

#include "AmRtpStream.h"
#include "AmThread.h"
#include "AmRequest.h"
#include "AmEventQueue.h"

#include <string>
#include <vector>
#include <queue>
#include <map>
using std::string;
using std::vector;
using std::queue;
using std::map;
using std::pair;

class AmStateFactory;
class AmDialogState;

class AmSessionEvent: public AmEvent
{
public:
    AmRequestUAS request;

    enum EventType { Refer,Notify,ReInvite };
    AmSessionEvent(EventType event_type, const AmRequestUAS& request);
};

/**
 * The session class.
 * It implements the behavior of the 
 * answer machine (play, record then email).
 * The session is identified by Call-ID, From-Tag and To-Tag.
 * @todo Use email templates instead hard coded ones...
 */
class AmSession : public AmThread
{
    SdpPayload   payload;
    bool         first_negotiation;

    /** @see AmThread::run() */
    void run();

    /** @see AmThread::on_stop() */
    void on_stop();

    /** @see AmThread::on_stop() */
    void onIdle();

protected:

    AmDialogState* dialog_st;

public:

    AmRtpStream         rtp_str;
    auto_ptr<AmRequest> req;

    AmCondition<bool> sess_stopped;

    struct Exception {
	int code;
	string reason;
	Exception(int c, string r) : code(c), reason(r) {}
    };

    /** 
     * Session constructor.
     * @param _req request which initiated the session 
     * @param _cb  callbacks of the application.
     */
    AmSession(AmRequest* _req, AmDialogState* _d_st);
    virtual ~AmSession();

    /** Gets the Session's call ID */
    const string& getCallID();

    /** Gets the Session's remote tag */
    const string& getFromTag();

    /** Gets the Session's local tag */
    const string& getToTag();

    /** Gets the dialog state */
    AmDialogState* getDialogState();

    /** Gets the current RTP payload */
    const SdpPayload& getPayload();

    /** Gets the port number of the remote part of the session */
    int getRPort();

    /** handle SDP negotiation: only for INVITEs & re-INVITEs */
    void negotiate(AmRequest* request);

    /**
     * Destroy the session.
     * It causes the session to be erased from the active session list
     * and added to the dead session list.
     * @see AmSessionContainer
     */
    void destroy();

    /**
     * Creates a new Id which can be used within sessions.
     */
    static string getNewId();
};


/**
 * Centralized session container.
 * This is the register for all active and dead sessions.
 * If has a deamon which wakes up only if it has work. 
 * Then, it kills all dead sessions and try to go to bed 
 * (it cannot sleep if one or more sessions are still alive).
 */
class AmSessionContainer : public AmThread
{
    static AmSessionContainer* _SessionContainer;

    // some typedefs ...
    typedef pair<string,AmSession*> SessionTableEntry;

    typedef vector<SessionTableEntry> SessionTable;
    typedef vector<SessionTableEntry>::iterator SessionTableIter;

    typedef map<string,SessionTable> SessionMap;
    typedef map<string,SessionTable>::iterator SessionMapIter;

    typedef pair<SessionMapIter,SessionTableIter> SessionSearchRes;

    typedef queue<AmSession*> SessionQueue;

    /** Container for active sessions */
    SessionMap a_sessions;
    /** Mutex to protect the active session container */
    AmMutex    as_mut;

    /** Container for dead sessions */
    SessionQueue d_sessions;
    /** Mutex to protect the dead session container */
    AmMutex    ds_mut;

    /** the daemon only runs if this is true */
    AmCondition<bool> _run_cond;

    /** We are a Singleton ! Avoid people to have their own instance. */
    AmSessionContainer();

    /** 
     * Search for a session in the active session container.
     * @param hash_str Call-ID + From-tag
     * @param sess_key [in,out]To-tag
     * @return the search results...
     */
    SessionSearchRes findSession(const string& hash_str, string& sess_key);

    /**
     * Search the container for a session coresponding to hash_str and sess_key.
     * @param hash_str  Call-ID + From-tag
     * @param sess_key  To-tag
     * @param lock      Should the container get locked while searching ?
     * @return the session related to hash_str & sess_key 
     *         or NULL if none has been found.
     */
    AmSession* getSession(const string& hash_str, string& sess_key);
    /**
     * Destroys a session.
     * Processes the result value of findSession().
     * @param sess_pair see findSession().
     * @param sessionsAreLocked is the active session container locked ?
     */
    void destroySession(pair<SessionMapIter,SessionTableIter>& sess_pair, bool sessionsAreLocked);
    /**
     * Tries to stop the session and queue it destruction.
     */
    void stopAndQueue(AmSession* s);

    /** @see AmThread::run() */
    void run();
    /** @see AmThread::on_stop() */
    void on_stop();

 public:
    static AmSessionContainer* instance();

    /**
     * Creates a new session.
     * @param req local request
     * @return a new session or NULL on error.
     */
    AmSession* createSession(const string& app_name, AmRequest* req);
    /**
     * Adds a session to the container.
     * @param hash_str Call-ID + remote-tag
     * @param sess_key local-tag
     * @return true if the session is new within the container.
     */
    bool addSession(const string& hash_str,const string& sess_key,
		    AmSession* session);
    /** 
     * Constructs a new session and adds it to the active session container. 
     * @param hash_str Call-ID + remote-tag
     * @param sess_key local-tag
     * @param req client's request
     */
    void startSession(const string& hash_str,string& sess_key, AmRequestUAS* req);

    /** @return false if session doesn't exist */
    bool postEvent(const string& hash_str,string& sess_key,
		   AmSessionEvent* event);

    /**
     * Seek & destroy session.
     * @param hash_str Call-ID + From-tag
     * @param sess_key [in,out]To-tag
     * @return true if session exists
     */
    bool sadSession(const string& hash_str, string& sess_key); // seek & sestroy session
    /**
     * Detroys a session.
     */
    void destroySession(AmSession* s);
};

#endif

// Local Variables:
// mode:C++
// End:

