/*
 * $Id: AmAudioIsdn.cpp,v 1.2 2003/11/16 19:17:55 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmAudioIsdn.h"
#include "log.h"

/** isdn audio is mono */
#define ISDNAUDIO_CHANNELS 1
/** isdn audio sample rate is 8000 */
#define ISDNAUDIO_RATE 8000
/** one sample per frame */
#define ISDNAUDIO_SAMPLE 1
/** codec to be used while communicating with isdn device (u-law) */
#define ISDNAUDIO_CODEC_ULAW 1
#define ISDNAUDIO_CODEC_ALAW 2

/** Creates an converter in <B>one</B> direction. Two instances of 
 * this class are needed for a duplex connection, one with
 * fromIsdn=false and one with fromIsdn=true 
 */
AmAudioIsdn::AmAudioIsdn (IsdnDevice *idev, bool fromIsdn) 
	: AmAudio(), m_idev(idev)
{
	if (fromIsdn) {
		/* connection isdn -> rtp, out will be set in duplex() */
		in.reset(new AmAudioIsdnFormat(ISDNAUDIO_CODEC_ULAW));
	} else {
		/* connection rtp -> isdn, in will be set in duplex() */
		out.reset(new AmAudioIsdnFormat(ISDNAUDIO_CODEC_ULAW));
	}
	#ifdef USE_CAPI20
	m_ccon = NULL;
	#endif
}

#ifdef USE_CAPI20
AmAudioIsdn::AmAudioIsdn (IsdnCapiConnection *ccon, bool fromIsdn) 
	: AmAudio(), m_ccon(ccon), m_idev(NULL)
{
	if (fromIsdn) {
		/* connection isdn -> rtp, out will be set in duplex() */
		in.reset(new AmAudioIsdnFormat(ISDNAUDIO_CODEC_ALAW));
	} else {
		/* connection rtp -> isdn, in will be set in duplex() */
		out.reset(new AmAudioIsdnFormat(ISDNAUDIO_CODEC_ALAW));
	}
}
#endif

/** Destructor. Unused by now. */
AmAudioIsdn::~AmAudioIsdn () {}

/** Make the codec id to be used available to others. 
 * This method is required by the sems api. */
int AmAudioIsdnFormat::getCodecId() {
	return (m_type);
}	

/** Close this AmAudio inherited device. Closing is not neccessary here, but
 * the sems api requires this method, so it's noop.
 */
void AmAudioIsdn::close() {
}

/** Get size bytes from the isdn audio channel. This method is called every 
 * time the audio converter inside sems needs more input. Reading and filtering
 * is actually done in IsdnDevice. */
int AmAudioIsdn::streamGet(unsigned int user_ts, unsigned int size) {
	if (m_idev != NULL) { /* i4l connection */
		int ret = m_idev->getVoiceStream( (unsigned char*)samples, size);
		return ret;
	} else {
		#ifdef USE_CAPI20
		return (m_ccon->getVoiceStream( (unsigned char*)samples, size));
		#else
		ERROR("no capi support, should not get here (01)\n");
		return -1;
		#endif
	}	
}	

/** Put size bytes to the isdn audio channel. The converter inside sems pushes
 * audio data using this method. \todo Filtering should take place in IsdnDevice. */
int AmAudioIsdn::streamPut(unsigned int user_ts, unsigned int size) {
	if (m_idev != NULL) { /* i4l connection */
	
		/* TODO: put all this code into IsdnDevice */
		int write_count;
		unsigned char *write_me = (unsigned char*) malloc (size * 2);
		if (write_me == NULL) {
			ERROR("out of memory\n");
			exit(-1);
		}	
		
		/* need to filter special character 0x10, see isdn_audio(4) for details */
		unsigned char *foo = write_me;
		for (unsigned int i=0; i<size; i++) {
			*(foo++) = *(samples+i);
		/* insert another 0x10 after each 0x10 */
			if (*(samples+i) == 0x10) 
				*(foo++) = 0x10;
		}
		write_count = foo - write_me;
		// *foo = '\0';
		int ret;
		/* actually write the data to the isdn device */
		ret = m_idev->putVoiceStream(write_me, write_count);
		free(write_me);
		return ret; 
	} else { /* capi connection */
		#ifdef USE_CAPI20
		return (m_ccon->putVoiceStream( (unsigned char*)samples, size));
		#else
		ERROR("no capi support, should not get here (02)\n");
		return -1;
		#endif
	}	
}	

/** ISDN audio format, used for duplex connection. This subclass just
 * sets the appropriate members of AmAudioFormat and initializes the 
 * codec.
 */
AmAudioIsdnFormat::AmAudioIsdnFormat (int type) : m_type(type) {
	/* set audio parameters for isdn4linux format (ulaw selected) */
	channels = ISDNAUDIO_CHANNELS;
	rate = ISDNAUDIO_RATE;
	sample = ISDNAUDIO_SAMPLE;
	/* initialize the codec */
	codec = getCodec();
	initCodec();
	if (!codec) 
		INFO("ERROR: could not initialize isdn codec!\n");
	/* this is probably to state: we have a running codec */
	h_codec=1;
}	


