/*
 * $Id: IsdnCall.cpp,v 1.5 2004/04/05 12:06:23 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "log.h"
#include "AmSession.h"
#include "AmAudioIsdn.h"

#include "IsdnCall.h"
#include "IsdnCallManager.h"
#include "IsdnDevice.h"
#include "globals.h"
#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "IsdnDeviceOutgoing.h"
#include "IsdnDeviceIncoming.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"

IsdnCall::IsdnCall() {
	m_pstnstate = CALLSTATE_IDLE;
	m_sipstate = CALLSTATE_IDLE;
	m_sipcode = 0;
	pthread_mutex_init(&setup_m, NULL);
	*m_fromnumber = '\0';
	*m_tonumber = '\0';
	*m_fromSip = '\0';
	*m_toSip = '\0';
	m_calldetail = callManager->addCall(this);
}

IsdnCall::~IsdnCall() {
#ifdef USE_CAPI20
	unlock();
	if (m_backendtype == CONNECTIONTYPE_CAPI) {
		delete ((IsdnCapiConnection*)m_pstndevice);
	}	
#endif

	callManager->diedCall(this, m_calldetail);

}

void IsdnCall::lock() {
	pthread_mutex_lock(&setup_m);
}

void IsdnCall::unlock() {
	pthread_mutex_unlock(&setup_m);
}

void IsdnCall::setRequest(AmRequest* req) {
	m_req = req;
}

void IsdnCall::setSession(AmSession* ses) {
	m_session = ses;
}

void IsdnCall::setCommand(AmCmd cmd) {
	m_cmd = cmd;
}

char* IsdnCall::getFromNumber() {
	return m_fromnumber;
}

char* IsdnCall::getFromSip() {
	return m_fromSip;
}

char* IsdnCall::getToNumber() {
	return m_tonumber;
}

char* IsdnCall::getToSip() {
	return m_toSip;
}



void IsdnCall::placeCallFromSIP(AmCmd cmd) {
	int device;

	m_req = NULL;
	m_cmd = cmd;
	m_pstndevice = NULL;
	m_fromPSTN = false;
	
	/* SIP uris are of no interest here */
	strcpy(m_fromSip, cmd.from.c_str());
	strcpy(m_toSip, cmd.to.c_str());
	
	DBG("******************* ***************** storing from/tosip: >%s</>%s<\n", m_fromSip, m_toSip);
	
	IsdnUtils::determineNumbers(m_cmd, m_tonumber, m_fromnumber, MAX_NUM_LEN);
	device = isdn_config->getDeviceForCall(m_tonumber, m_fromnumber);
	switch (device) {
		case DEVICE_NO_DEVICE : 
			ERROR("isdngw: no rule to handle call from %s to %s\n", m_fromnumber, m_tonumber); 
			throw AmSession::Exception(500,"isdngw: no device configured to handle this call");
			break;
		case DEVICE_I4L : 
			IsdnDeviceOutgoing *cur;
			/* get and lock a free device */
			cur = (IsdnDeviceOutgoing *)IsdnDeviceOutgoing::getLockedDevice();
			if (cur == NULL) {
				throw AmSession::Exception(500,"isdngw: sorry, no outgoing i4l device available");
			}	
			cur->setCall(this);
			m_backendtype = CONNECTIONTYPE_I4L;
			m_pstndevice = cur;
			break;
		default: 
			#ifdef USE_CAPI20
			IsdnCapiConnection *ccon = new IsdnCapiConnection(device);
			ccon->setCall(this);
			/*strcpy(ccon->m_from, m_fromnumber);
			strcpy(ccon->m_to, m_tonumber);
			*/
			m_backendtype = CONNECTIONTYPE_CAPI;
			m_pstndevice = ccon;
			#else
			ERROR("isdngw: no CAPI support available, compile with -DUSE_CAPI20\n");
			throw AmSession::Exception(500,"isdngw: sorry, internal error (capi support disabled)");
			#endif
	}
	char b[100];
	if (device == DEVICE_I4L)
		sprintf(b, "isdn4linux device");
	else
		sprintf(b, "capi controller %i", device);
	if ((m_fromnumber == NULL) || (strlen(m_fromnumber) == 0))
		strcpy(m_fromnumber, "0");
	INFO("isdngw: using %s to handle call from %s to %s\n", b, m_fromnumber, m_tonumber);
	callManager->updateCallParameters(this, m_calldetail);
}

void IsdnCall::placeCallFromPSTN(char* fromnumber, char* tonumber, int backend, void* device) {
	m_req = NULL;
	m_pstndevice = device;
	m_fromPSTN = true;
	m_backendtype = backend;
	strcpy(m_fromnumber, fromnumber);
	strcpy(m_tonumber, tonumber);
	callManager->updateCallParameters(this, m_calldetail);
}

// helper
char* textForState(int s) {
	if (s == CALLSTATE_IDLE) return "IDLE";
	if (s == CALLSTATE_CONNECTING) return "CONNECTING";
	if (s == CALLSTATE_CONNECTED) return "CONNECTED";
	if (s == CALLSTATE_DISCONNECTED) return "DISCONNECTED";
	return "illegal value";
}	

/* short info on the state machine used here:
 * every state is defined by the PSTN-state and the SIP-state.
 * Every state transition is checked against a positive list, if
 * this partitcular state transition is not defined, you can call
 * goto unhandled, resulting in a printout of missing state transition.
 */
void IsdnCall::changeSIPState(int newstate, int code, const char* reason) {
	int ret = OK;
	int oldstate = m_sipstate;
	int f1, f2, f3, f4;
	char buf[255];
	sprintf(buf, "CALL: SIP  state (%s): from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound ":"outbound", textForState(m_sipstate), textForState(m_pstnstate), textForState(newstate), textForState(m_pstnstate));
	DBG("%s", buf);
	callManager->updateCall(this, m_calldetail, buf);
	f1 = m_sipstate;
	f2 = m_pstnstate;
	f3 = newstate;
	f4 = m_pstnstate;
	
	m_sipstate = newstate;
	switch (oldstate) {
		case CALLSTATE_IDLE: 
			if (m_fromPSTN) { /* inbound logic */
				if (m_pstnstate != CALLSTATE_IDLE) {
					goto unhandled;
				} else if (newstate == CALLSTATE_CONNECTING) {
					//DBG("SIP state now connecting\n");
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (m_pstnstate != CALLSTATE_IDLE) {
					goto unhandled;
				}	
				if (newstate == CALLSTATE_CONNECTING) { /* (1) IDLE|IDLE -> CONNECTING|IDLE */
					lock(); /* lock before call to placeCall() */
					if (m_backendtype == CONNECTIONTYPE_I4L) {
						((IsdnDeviceOutgoing*)m_pstndevice)->placeCall(m_req, m_tonumber, m_fromnumber);
					} else {
						#ifdef USE_CAPI20
						ret = ((IsdnCapiConnection*)m_pstndevice)->placeCall(m_req, m_tonumber, m_fromnumber);
						#endif
					}	
					if (ret == FAIL) {
						ERROR("internal error, could not place call\n");
						changePSTNState(CALLSTATE_DISCONNECTED);  /* (2) IDLE => DISCONNECTED */
					}	else {
						changePSTNState(CALLSTATE_CONNECTING);    /* (2) IDLE => CONNECTING */
					}	
				} else {
					goto unhandled;
				}
			}	
			break;
		case CALLSTATE_CONNECTING: 
			if (m_fromPSTN) { /* inbound logic */
				if (newstate == CALLSTATE_CONNECTED)	{
					DBG("SIP side connected, now connecting PSTN side\n");
					changePSTNState(CALLSTATE_CONNECTING);
				} else if (newstate == CALLSTATE_DISCONNECTED) {
					if (m_pstnstate == CALLSTATE_IDLE) 
						changePSTNState(CALLSTATE_DISCONNECTED);
					else	
						AmRequestUAC::cancel(m_cmd);
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (newstate == CALLSTATE_CONNECTED) {
					if (m_pstnstate != CALLSTATE_CONNECTED)
						goto unhandled;
					connectAudio();
				} else if (newstate == CALLSTATE_DISCONNECTED) {
					if (m_pstnstate != CALLSTATE_DISCONNECTED) {
						changePSTNState(CALLSTATE_DISCONNECTED);
					} 
					if (code != 0) {
						DBG("storing reply to request: %i %s\n", code, reason);
						m_sipcode = code;
						strncpy(m_sipreason, reason, 100);
					} else {
						DBG("storing call terminated as stored exception\n");
						m_sipcode = 487;
						strncpy(m_sipreason, "isdngw: call terminated", 100);
					}
				} else	
					goto unhandled;	
			}	
			break;
		case CALLSTATE_CONNECTED: 
			if (m_fromPSTN) { /* inbound logic */
				if (newstate != CALLSTATE_DISCONNECTED) {
					goto unhandled;
				} else if (m_pstnstate == CALLSTATE_CONNECTED) {
					DBG("SIP side ended the call\n");
					changePSTNState(CALLSTATE_DISCONNECTED);
				} else if (m_pstnstate == CALLSTATE_DISCONNECTED) {
					DBG("PSTN side ended the call\n");
					m_req->bye();
				} else if (m_pstnstate == CALLSTATE_CONNECTING) {
					DBG("PSTN side could not connect just when SIP accepted the call\n");
					changePSTNState(CALLSTATE_DISCONNECTED);
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (newstate == CALLSTATE_DISCONNECTED) {
					if (m_pstnstate == CALLSTATE_CONNECTED) {
						m_req->bye();
						changePSTNState(CALLSTATE_DISCONNECTED);
					} else if (m_pstnstate == CALLSTATE_DISCONNECTED) {
							if (code != 0) {
								m_req->reply(code, reason);
							} else {
								m_req->bye();
							}	
					} else {
						goto unhandled;
					}	
				}	else
					goto unhandled;
			}		
			break;
		case CALLSTATE_DISCONNECTED: 
			if (newstate == CALLSTATE_DISCONNECTED) {
				if (m_pstnstate == CALLSTATE_DISCONNECTED) {
					//DBG("SIP: both states disconnected, nothing to do\n");
				} else 
					goto unhandled;
			} else		
				goto unhandled;
			break;
		default: 
			ERROR("unknown SIP state %i before change to %i\n", m_sipstate, newstate);
	}		

	sprintf(buf, "DONE: SIP  state (%s): from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound ":"outbound", textForState(f1), textForState(f2), textForState(f3), textForState(f4));
	DBG("%s", buf);
/* commented out: cannot make shure the current object still exists, maybe the session cleaner came around... */
/*	callManager->updateCall(this, m_calldetail, buf); */

	return;
	unhandled:
		sprintf(buf, "unhandled SIP  state transition (%s) from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound":"outbound", textForState(f1), textForState(f2), textForState(f3), textForState(f4));
		callManager->updateCall(this, m_calldetail, buf);
		ERROR("%s", buf);
		exit(1);
}

void IsdnCall::changePSTNState(int newstate, int code, const char* reason) {
	char buf[255];
	int oldstate = m_pstnstate;
	int f1, f2, f3, f4;
	sprintf(buf, "CALL: PSTN state (%s): from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound ":"outbound", textForState(m_sipstate), textForState(m_pstnstate), textForState(m_sipstate), textForState(newstate));
	DBG("%s", buf);
	callManager->updateCall(this, m_calldetail, buf);
	/* store for possible errors */
	f1 = m_sipstate;
	f2 = m_pstnstate;
	f3 = m_sipstate;
	f4 = newstate;
	m_pstnstate = newstate;
	switch (oldstate) {
		case CALLSTATE_IDLE: 
			if (m_fromPSTN) { /* inbound logic */
				if (newstate == CALLSTATE_CONNECTING) {
					int ret = OK;
					DBG("now connecting PSTN side\n");
					if (m_backendtype == CONNECTIONTYPE_I4L) {
						ret = ((IsdnDeviceIncoming*)m_pstndevice)->connectVoice();
					} else {
						#ifdef USE_CAPI20
						if (m_req == NULL) {
							ERROR("m_req is NULL!\n");
						}	
						((IsdnCapiConnection*)m_pstndevice)->onBeforeCallAccept(m_req);
						#endif
					}	
					if (ret == FAIL) {
						changeSIPState(CALLSTATE_DISCONNECTED);
					} else {
						DBG("connecting incoming call on pstn side success, changing state to CONNECTED\n");
						changePSTNState(CALLSTATE_CONNECTED);
					}	
				} else if (newstate == CALLSTATE_DISCONNECTED) { /* cancel */
					DBG("PSTN hangup before call was successfully setup\n");
					if (m_sipstate != CALLSTATE_DISCONNECTED)
						changeSIPState(CALLSTATE_DISCONNECTED);
					if (m_backendtype == CONNECTIONTYPE_I4L) {
						unlock();
						((IsdnDeviceIncoming*)m_pstndevice)->resetFailedIncomingDevice();
						((IsdnDeviceIncoming*)m_pstndevice)->reassignIncomingDevices();
					} else if (m_backendtype == CONNECTIONTYPE_CAPI) {
						/* revoke alert */
						#ifdef USE_CAPI20
						DBG("Revoke ALERT command.\n");
						((IsdnCapiConnection*)m_pstndevice)->hangup();
						#endif
					}
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (m_sipstate == CALLSTATE_DISCONNECTED) {
					if (m_backendtype == CONNECTIONTYPE_I4L) { /* SIP sent CANCEL ! */
						DBG("now setting hangup requested flag\n");
            ((IsdnDeviceOutgoing*)m_pstndevice)->setHangupRequested();
					} else {
						goto unhandled;
					}	
				} else if (m_sipstate != CALLSTATE_CONNECTING) {
					goto unhandled;
				} else 	if (newstate == CALLSTATE_CONNECTING) { /* (2) CONNECTING|IDLE -> CONNECTING|CONNECTING */
					if (m_backendtype == CONNECTIONTYPE_I4L) {
						int ret = ((IsdnDeviceOutgoing*)m_pstndevice)->connectVoice();
						unlock();
						if (ret == FAIL) {
							ERROR("could not connect voice for I4l outgoing connection\n");
						}
					}
					//DBG("PSTN changing from idle to connected\n");
					m_req->reply(180, "isgndw: ringing");
					lock(); /* wait until pstn is ready */
					unlock(); /* reset the lock */
					if (m_pstnstate != CALLSTATE_DISCONNECTED) {
						//DBG("done waiting: now change PSTN state to connected\n");
						changePSTNState(CALLSTATE_CONNECTED);
					}	else {
						if (m_fromPSTN) {
							//DBG("throwing 487!\n");
							throw AmSession::Exception(487, "call terminated");
						} else {
							//DBG("checking for stored Exception, because fromSIP\n");
							if (m_sipcode != 0) {
								//DBG("found stored exception, throwing: %i %s \n", m_sipcode, m_sipreason);
								throw AmSession::Exception(m_sipcode, m_sipreason);
							} else {
								//ERROR("isdngw: internal error, dont know which exception to send, sending 487 terminated!\n");
								throw AmSession::Exception(487, "isdngw: call terminated");
							}	
						}	
					}
				}	else if (newstate == CALLSTATE_DISCONNECTED) {
					if (code != 0) {
						//DBG("throwing exception!\n");
						throw AmSession::Exception(code,reason);
					} else {
						//DBG("throwing 500!\n");
						throw AmSession::Exception(500,"internal gateway error");
					}	
				} else 	
					goto unhandled;
			}		
			break;
		case CALLSTATE_CONNECTING: 
			if (m_fromPSTN) { /* inbound logic */
				if (m_sipstate == CALLSTATE_DISCONNECTED) {
					hangupPSTN();
					unlock();
				} else if (m_sipstate != CALLSTATE_CONNECTED) {
					goto unhandled;
				} else if (newstate == CALLSTATE_CONNECTED) {	
					if (m_backendtype == CONNECTIONTYPE_I4L) {
						DBG("terminating i4l inbound listener thread\n");
						((IsdnDeviceIncoming*)m_pstndevice)->terminateListener();
					}	
					// wait until capi connection is really up
/*					#ifdef USE_CAPI20
					if (m_backendtype == CONNECTIONTYPE_CAPI) {
						int retry = 0;
						while (retry++<10) {
								DBG("waiting\n");	
							if ( ((IsdnCapiConnection*)m_pstndevice)->getAudioConnected() )
								break;
							usleep(100000);	
						}	
					}
					#endif 
*/					
					DBG("now both sides connected, running connectAudio!\n");
					connectAudio();
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (newstate == CALLSTATE_CONNECTED) {
					if (m_sipstate != CALLSTATE_CONNECTED)
						return;
					else	
						goto unhandled;
					DBG("PSTN now connected\n");
				}	else if (newstate == CALLSTATE_DISCONNECTED) {
					if (m_sipstate == CALLSTATE_CONNECTING) {
						changeSIPState(CALLSTATE_DISCONNECTED, code, reason);
					} else {	
						hangupPSTN();
						unlock();
					}	
				} else 	
					goto unhandled;
			}		
			break;
		case CALLSTATE_CONNECTED:
			if (m_fromPSTN) { /* inbound logic */
				if (m_sipstate == CALLSTATE_DISCONNECTED) {
					DBG("SIP hung up, now hanging up PSTN\n");
					hangupPSTN();
				} else if (m_sipstate != CALLSTATE_CONNECTED) {
					goto unhandled;
				} else if (newstate == CALLSTATE_DISCONNECTED) {
					DBG("PSTN hangup detected\n");
					hangupPSTN();
					unlock();
					changeSIPState(CALLSTATE_DISCONNECTED);
				} else {
					goto unhandled;
				}	
			} else { /* outbound logic */
				if (newstate == CALLSTATE_DISCONNECTED) {
					if (m_sipstate == CALLSTATE_DISCONNECTED) {
						hangupPSTN();
						unlock();
					} else if (m_sipstate == CALLSTATE_CONNECTED) {
						hangupPSTN();
						unlock();
						changeSIPState(CALLSTATE_DISCONNECTED, code, reason);
					} else {
						goto unhandled;
					}	
				} else
					goto unhandled;
			}		
			break;
		case CALLSTATE_DISCONNECTED:
			if (newstate == CALLSTATE_CONNECTED) {
				DBG("pstn call setup failed (%i %s)\n", code, reason);
				if (m_sipstate == CALLSTATE_CONNECTING) {
					if (code != 0)
						changeSIPState(CALLSTATE_DISCONNECTED, code, reason);
					else
						changeSIPState(CALLSTATE_DISCONNECTED);
				} else 
					goto unhandled;
			} else if (newstate == CALLSTATE_DISCONNECTED) {
				if (m_sipstate == CALLSTATE_DISCONNECTED) {
					//DBG("both states disconnected, nothing to do\n");
				} else 
					goto unhandled;
			} else 		
				goto unhandled;
			break;
		default: 
			ERROR("unknown PSTN state %i before change to %i\n", m_pstnstate, newstate);
	}		

	sprintf(buf, "DONE: PSTN state (%s): from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound ":"outbound", textForState(f1), textForState(f2), textForState(f3), textForState(f4));
	DBG("%s", buf);
/* commented out: cannot make shure the current object still exists, maybe the session cleaner came around... */
/* 	callManager->updateCall(this, m_calldetail, buf); */
	

	return;
	unhandled:
		sprintf(buf, "unhandled PSTN state transition (%s) from (%s|%s) to (%s|%s)\n", m_fromPSTN?"inbound":"outbound", textForState(f1), textForState(f2), textForState(f3), textForState(f4));
		callManager->updateCall(this, m_calldetail, buf);
		ERROR("%s", buf);
		exit(1);
}

/* to place somewhere ;-) */
/*
	if (m_pstndevice->connectVoice() == FAIL) {
			throw AmSession::Exception(500,"isdngw: sorry, error intitiating voice transmission.");
		}	
*/
void IsdnCall::hangupPSTN() {
	//DBG("hangup pstn called\n");
	if (m_backendtype == CONNECTIONTYPE_I4L) {
		((IsdnDeviceOutgoing*)m_pstndevice)->hangup();
		((IsdnDeviceOutgoing*)m_pstndevice)->initialize();
		((IsdnDeviceOutgoing*)m_pstndevice)->unlock();
	} else {
		#ifdef USE_CAPI20
		DBG("PSTN hangup CAPI device\n");
		((IsdnCapiConnection*)m_pstndevice)->hangup();
		#endif
	}
}	

void IsdnCall::connectAudio() {
	if (m_backendtype == CONNECTIONTYPE_I4L) {
		/* takes i4l as input and rtp as output */
	 	AmAudioIsdn isdn2rtp((IsdnDevice*)m_pstndevice, true);
		/* takes rtp as input and i4l as output */
		AmAudioIsdn rtp2isdn((IsdnDevice*)m_pstndevice, false);
		if (m_session == NULL) 
			ERROR("m_session is NULL\n");
		else {
			callManager->updateCall(this, m_calldetail, "AUDIO CONNECTED\n");
			m_session->rtp_str.duplex(&isdn2rtp, &rtp2isdn);
			callManager->updateCall(this, m_calldetail, "AUDIO DISCONNECTED\n");
		}
		/* if done && inbound && pstn hup send bye to SIP */
		if (m_fromPSTN && m_sipstate != CALLSTATE_DISCONNECTED) {
			m_req->bye();
			changeSIPState(CALLSTATE_DISCONNECTED);
		}	
	} 
	#ifdef USE_CAPI20
	else if (m_backendtype == CONNECTIONTYPE_CAPI) {
		/* takes CAPI connection as input and rtp as output */
		AmAudioIsdn isdn2rtp((IsdnCapiConnection*)m_pstndevice, true);
		/* takes rtp as input and CAPI connection as output */
		AmAudioIsdn rtp2isdn((IsdnCapiConnection*)m_pstndevice, false);
		if (m_session == NULL) {
			ERROR("m_session is NULL\n");
		} else {
			callManager->updateCall(this, m_calldetail, "AUDIO CONNECTED\n");
			m_session->rtp_str.duplex(&isdn2rtp, &rtp2isdn);
			callManager->updateCall(this, m_calldetail, "AUDIO DISCONNECTED\n");
		}
	}
	#endif
	else {
		ERROR("unknown backend\n");
		return;
	}	
	//DBG("connectaudio done.\n");
}

bool IsdnCall::isDisconnected() {
	return (m_pstnstate == CALLSTATE_DISCONNECTED);
}	


/* reporting functions, used by isdncallmanager */

char* IsdnCall::getBackendAsText() {
	if (m_backendtype == CONNECTIONTYPE_CAPI)
		return "CAPI";
	else if (m_backendtype == CONNECTIONTYPE_I4L)
		return "I4L";
	else
		ERROR("bug!\n");
	return "ERROR";
}

char* IsdnCall::getSIPStateAsText() {
	return textForState(m_sipstate);
}

char* IsdnCall::getPSTNStateAsText() {
	return textForState(m_pstnstate);
}

const char* IsdnCall::getDirection() {
	return (m_fromPSTN?"PSTN2SIP":"SIP2PSTN");
}

char* IsdnCall::getFromAsText() {
	if (m_backendtype == CONNECTIONTYPE_I4L) {
		return "?";
	} else if (m_backendtype == CONNECTIONTYPE_CAPI) {
		#ifdef USE_CAPI20
		DBG("################### mfromnumber: %s : %s : \n", m_fromnumber, m_fromSip);
		return m_fromnumber;
		#endif
	} else {
		ERROR("BUG!\n");
	}	
	return "ERROR";
}

char* IsdnCall::getToAsText() {
	if (m_backendtype == CONNECTIONTYPE_I4L) {
		return "?";
	} else if (m_backendtype == CONNECTIONTYPE_CAPI) {
		#ifdef USE_CAPI20
		return m_tonumber;
		#endif
	} else {
		ERROR("BUG!\n");
	}	
	return ("ERROR");
}

char* IsdnCall::getDeviceAsText() {
	static char buf[3];
	if (m_backendtype == CONNECTIONTYPE_I4L) {
		return "0";
	} else if (m_backendtype == CONNECTIONTYPE_CAPI) {
		#ifdef USE_CAPI20
		sprintf(buf, "%i", ((IsdnCapiConnection*)m_pstndevice)->getController());
		#endif
		return buf;
	} else {
		ERROR("bug! unknown conn type\n");
	}	
	return ("ERROR");
}



