/*
 * $Id: IsdnCallManager.cpp,v 1.6 2004/04/04 14:11:28 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "log.h"
#include "AmSession.h"
#include "AmUtils.h"

#include "globals.h"
#include "IsdnCallManager.h"
#include "IsdnDevice.h"
#include "IsdnConfiguration.h"
#include "IsdnDeviceIncoming.h"
#include "IsdnDevice.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"
#include "IsdnCall.h"

/* globally available callManager instance */
IsdnCallManager* callManager = NULL;

IsdnCallManager::IsdnCallManager() {
	m_callDetails = NULL;
	pthread_mutex_init(&details_m, NULL);
}

IsdnCallManager::~IsdnCallManager() {
	if (m_callDetails != NULL)
		delete m_callDetails;
	pthread_mutex_unlock(&details_m);
	pthread_mutex_destroy(&details_m);
}

void IsdnCallManager::instantiate() {
	if (callManager == NULL) {
		callManager = new IsdnCallManager();
		callManager->init();
	}
}

void IsdnCallManager::init() {
}

IsdnCall* IsdnCallManager::placeCallFromPSTN(char* fromnumber, char* tonumber, int backend, void* device) {
	IsdnCall* call = new IsdnCall();
	// TODO: try block to free call if unsuccessful
	call->placeCallFromPSTN(fromnumber, tonumber, backend, device);
	return call;
}

IsdnCall* IsdnCallManager::placeCallFromSIP(AmCmd cmd) {
	IsdnCall* call = new IsdnCall();
	// TODO: try block to free call if unsuccessful
	call->placeCallFromSIP(cmd);
	return call;
}

IsdnCall* IsdnCallManager::changeSIPState(AmCmd cmd, int newstate) {
	IsdnCall* call;
	// find right isdncall, call changeSIPState there
	// this is only called for oninvite for inbound call
	IsdnDeviceIncoming *matching_device = IsdnDeviceIncoming::getPendingCallDevice(cmd.user.c_str());
	if (matching_device != NULL) {
		/* found matching i4l device! */
		call = matching_device->getCall();
	} else {
		#ifdef USE_CAPI20
		IsdnCapiConnection* in_ccon = capiManager->getPendingCallConnection(cmd.user.c_str());
		if (in_ccon != NULL) {
			/* found matching cconnection */
			call = in_ccon->getCall();
		}	else {
		#endif
		ERROR("isdngw: internal error, no matching inbound call found\n");
		throw AmSession::Exception(500,"isdngw: internal error, no matching inbound call found\n");
		#ifdef USE_CAPI20
		}
		#endif
	}
	call->changeSIPState(newstate);
	call->setCommand(cmd);
	return call;
}

/* obsolete: void IsdnCallManager::changeSIPState(AmRequest *req, int newstate) {

}
*/

#define MAX_DETAILIDS 100000
int IsdnCallManager::getNewId() {
	int ret;
	pthread_mutex_lock(&details_m);
	if (m_idcounter > MAX_DETAILIDS)
		m_idcounter = 0;
	while (getDetailsForId(++m_idcounter) != NULL);
	ret = m_idcounter;
	pthread_mutex_unlock(&details_m);
	return ret;
}

CallDetail* IsdnCallManager::addCall(IsdnCall* call) {
	DBG("new Call@%p for reporting app registered\n", call);
	/* create new callDetail object and attach to list */

	CallDetail* newdetail = new CallDetail();

	newdetail->next = NULL;
	newdetail->history = NULL;
	newdetail->id = getNewId();

	newdetail->fromuri = NULL;
	newdetail->touri = NULL;
	newdetail->fromnumber = NULL;
	newdetail->tonumber = NULL;
	newdetail->device = NULL;

	pthread_mutex_lock(&details_m);
	/*  not insert in list right now */
	CallDetail* cur = m_callDetails;
	if (cur == NULL) {
		m_callDetails = newdetail;
		pthread_mutex_unlock(&details_m);
		return newdetail;
	}	

	while (cur->next != NULL)
			cur = cur->next;
	cur->next = newdetail;
	
	pthread_mutex_unlock(&details_m);
	return newdetail;
}

void IsdnCallManager::updateCallParameters(IsdnCall* call, CallDetail* newdetail) {
	DBG("updateCallParameters!\n");
	// first fill a fields with valid information
	newdetail->fromnumber = (char*) malloc (strlen(call->getFromNumber())+1);
	if (newdetail->fromnumber == NULL) 
		ERROR("no memory for calldetail->fromnumber\n");
	else	
		strcpy(newdetail->fromnumber, call->getFromNumber());
	newdetail->tonumber = (char*) malloc (strlen(call->getToNumber())+1);
	if (newdetail->tonumber == NULL) 
		ERROR("no memory for calldetail->tonumber\n");
	else	
		strcpy(newdetail->tonumber, call->getToNumber());
	
	newdetail->fromuri = (char*) malloc (strlen(call->getFromSip())+1);
	if (newdetail->fromuri == NULL) 
		ERROR("no memory for calldetail->fromuri\n");
	else	
		strcpy(newdetail->fromuri, call->getFromSip());
	newdetail->touri = (char*) malloc (strlen(call->getToSip())+1);
	if (newdetail->touri == NULL) 
		ERROR("no memory for calldetail->touri\n");
	else	
		strcpy(newdetail->touri, call->getToSip());
	
	newdetail->device = (char*) malloc (strlen(call->getDeviceAsText())+1);
	if (newdetail->device == NULL) 
		ERROR("no memory for calldetail->device\n");
	else	
		strcpy(newdetail->device, call->getDeviceAsText());
	
	newdetail->backend = call->getBackendAsText();
	strcpy(newdetail->direction, call->getDirection());


	DBG("****** ******* ******** updatecall: *   * *** ** *** ** *: fromnumber/tonumber: %s/%s fromuri >%s< touri >%s<\n", newdetail->fromnumber, newdetail->tonumber, newdetail->fromuri, newdetail->touri);
}

void IsdnCallManager::diedCall(IsdnCall* call, CallDetail* callDetail) {
	DBG("Call object deleted, dumping details of call\n");
	dumpDetails(callDetail);
}

void IsdnCallManager::updateCall(IsdnCall* call, CallDetail* callDetail, char* entry) {
	/* report a status change */
	History* newhistory = new History();
	gettimeofday(&newhistory->timestamp, NULL);
	newhistory->description = (char*) malloc(strlen(entry)+1);
	if (newhistory->description == NULL) 
		ERROR("no memory to store call detail history entry\n");
	sprintf(newhistory->description, "%s", entry); 
	newhistory->next = NULL;
	History* cur = callDetail->history;
	if (cur == NULL) { /* no history yet */
		callDetail->history = newhistory;
	} else { /* append */
		while (cur->next != NULL)
			cur = cur->next;
		cur->next = newhistory;
	}	
}

void IsdnCallManager::dumpDetailsOfCalls() {
	int i = 0;
	for (CallDetail* cur = m_callDetails; cur!=NULL; cur = cur->next, i++) {
		dumpDetails(cur);
	}	
}

/* recursively free history */
void freeHistory (History* hist) {
	if (hist->next != NULL)
		freeHistory(hist->next);
	if (hist == NULL)
		return;
	free(hist->description);
	delete hist;
}	

/* returns CallDetails for a given id, NULL on non existing ids */
CallDetail* IsdnCallManager::getDetailsForId(int id) {
	CallDetail* cur = m_callDetails;
	while (cur != NULL) {
		if (cur->id == id) 
			return cur;
		cur = cur->next;
	}	
	return NULL;
}

void IsdnCallManager::dumpDetails(CallDetail* calldetail){
	if (calldetail == NULL) {
		DBG("will not dump NULL details...\n");
		return;
	}	
	History* cur = calldetail->history;
	DBG("dumping details for detail record@%p\n", calldetail);

	DBG("#%i: Call from %s to %s using device %s from backend %s\n", calldetail->id, calldetail->fromnumber, calldetail->tonumber, calldetail->device, calldetail->backend);
	if (calldetail->history == NULL) {
		DBG("  -> NO HISTORY\n");
	} else {
		do {
			/* DBG("  -> %u: %s (PSTN connected? %s)\n", cur->timestamp.tv_sec, cur->description, cur->pstnConnected?"yes":"no"); */
			DBG("  -> %u: %s", (unsigned) cur->timestamp.tv_sec, cur->description);
			cur = cur->next;
		} while (cur != NULL);
			
	}
/* 	freeHistory(calldetail->history);
	delete calldetail; */
}

void IsdnCallManager::fifo_listcalls(FILE* reply) {
	int number = 0;
	DBG("fifo_listcalls called\n");
	CallDetail* cur = m_callDetails;
	fprintf(reply, "OK\n");
	while (cur!=NULL) {
		fprintf(reply, "%i,%s,%s,%s,%s,%s,%s,%s\n", cur->id, cur->direction, cur->backend, cur->device, cur->fromnumber, cur->tonumber, cur->fromuri, cur->touri);
		number++;
		cur=cur->next;
	}
	fprintf(reply, "\n");
}

void IsdnCallManager::fifo_calldetails(FILE* reply, char* parameter) {
	int wantedid;
	CallDetail* cur;
	DBG("fifo_calldetails called (%s)\n", parameter);
	if (sscanf(parameter, "%i", &wantedid) != 1) {
		ERROR("invalid parameter %s\n", parameter);
		fprintf(reply, "ERROR\ninvalid parameter %s\n\n", parameter);
	}	else {
		cur = getDetailsForId(wantedid);
		if (cur == NULL) {
			ERROR("no call with id=%i available in detail records\n", wantedid);
			fprintf(reply, "ERROR\ninvalid parameter %i, no such record.\n\n",wantedid);
		} else {
			History* hist = cur->history;
			if (hist == NULL) {
				fprintf(reply, "ERROR\nno history available\n\n");
			} else {
				fprintf(reply, "OK\n");
				do {
				fprintf(reply,"%u: %s", (unsigned) hist->timestamp.tv_sec, hist->description);
				hist = hist->next;
				} while (hist != NULL);
				fprintf(reply, "\n");
			}
		}
	}	
}

int IsdnCallManager::deleteDetails(CallDetail* detail) {
	CallDetail* cur;
	if (detail == NULL)
		return FAIL;
	cur = m_callDetails;
#define f(x) if (x!=NULL)free(x);
	f(detail->tonumber);
	f(detail->touri);
	f(detail->fromnumber);
	f(detail->fromuri);
	f(detail->device);
	if (detail->history != NULL)
		freeHistory(detail->history);
	if (cur == detail) {
		m_callDetails = cur->next;	
		delete cur;
	} else {	
		while ( (cur->next != detail) && (cur->next != NULL))
			cur = cur->next;
		if (cur->next == NULL) {
			ERROR("internal error, could not delete calldetail at %p\n", detail);
			return FAIL;
		}	
		cur->next = detail->next;
	}	
	return OK;
}

void IsdnCallManager::fifo_deletedetails(FILE* reply, char* parameter) {
	int wantedid;
	CallDetail* cur;
	DBG("fifo_deletedetails called (%s)\n", parameter);
	if (sscanf(parameter, "%i", &wantedid) != 1) {
		ERROR("invalid parameter %s\n", parameter);
		fprintf(reply, "ERROR\ninvalid parameter %s\n\n", parameter);
	}	else {
		cur = getDetailsForId(wantedid);
		if (cur == NULL) {
			ERROR("no call with id=%i available in detail records\n", wantedid);
			fprintf(reply, "ERROR\ninvalid parameter %i, no such record.\n\n",wantedid);
		} else {
			if (deleteDetails(cur) == OK)
				fprintf(reply, "OK\n\n");
			else
				fprintf(reply, "ERROR\n\n");
		}
	}	
}

void IsdnCallManager::fifo_dumpdevices(FILE* reply) {
	DBG("fifo_dumpdevices called\n");
	fprintf(reply, "OK\n");
	isdn_config->fifo_dumpdevices(reply);	
	fprintf(reply, "\n");
}


/* web functions: 

  0.2
	isdninfo.<functionname>
	path/to/replyfifo
	parameter1
	parameter2
	parameterN..
	<CR>

	known functions: 
	listcalls, no parameters
	calldetails, parameter1: call-detail-id
	deletedetails: parameter1: call-detail-id

	*/
int IsdnCallManager::execute(FILE* stream, const string& s) {
	#define MAX_PARAMETER 10
	string line;
	char buf[MAX_LINE_SIZE];
	char parameter[MAX_PARAMETER][MAX_LINE_SIZE];
	int parametercount;
	int count = 0;
	string replyfifoname;
	FILE* replyfifo;
	bool error = false;
	int retrycount = 20;
	//DBG("FIFO command called %s\n", s.c_str());

	if ( (s == FIFO_COMMAND_LISTCALLS) || (s == FIFO_COMMAND_DUMPDEVICES)) {
		parametercount = 0;
	} else if ((s == FIFO_COMMAND_CALLDETAILS) || (s == FIFO_COMMAND_DELETEDETAILS)) {
		parametercount = 1;
	} else {
		ERROR("unknown FIFO command for isdngw: %s\n", s.c_str());
		parametercount = -1;
		error = true;
	}	

	READ_FIFO_PARAMETER(stream, replyfifoname, buf, MAX_LINE_SIZE);
	if (replyfifoname == "") {
		ERROR("no reply fifo filename specified, expected as first parameter\n");
		replyfifo = NULL;
	}
	
	again:
	int filedes = open (replyfifoname.c_str(), O_NONBLOCK | O_WRONLY);
	if (filedes == -1) {
		if (errno == 6) {
			usleep(100000);
			if (retrycount-- > 0)
				goto again;
		}	
		ERROR("could not open reply fifo %s for writing, reason %s\n", replyfifoname.c_str(), strerror(errno));
			
		replyfifo = NULL;
	} else {
		replyfifo = fdopen(filedes, "w");
		if (replyfifo == NULL) {
			ERROR("could not fdopen reply fifo, reason %s\n", strerror(errno));
		} else {
			if (error)
				fprintf(replyfifo, "ERROR\nunknown FIFO command\n\n");
		}	
	}	

	do {
		READ_FIFO_PARAMETER(stream, line, buf, MAX_LINE_SIZE);
		//DBG("got line: >>%s<<\n", line.c_str());
		if  ((line != "") && (count < MAX_PARAMETER - 1))
			strcpy(parameter[count++], line.c_str());
	} while (line != "");
	if ((parametercount != -1) && (count != parametercount)) { 
		ERROR("fifo function %s needs %i parameters but got %i\n", s.c_str(), parametercount, count);
		if (replyfifo != NULL)
			fprintf(replyfifo, "ERROR\nfifo function %s needs %i parameters but got %i\n\n", s.c_str(), parametercount, count);
	} else {
		if (s == FIFO_COMMAND_LISTCALLS) {
			if (replyfifo != NULL)
				fifo_listcalls(replyfifo);
		} else if (s == FIFO_COMMAND_CALLDETAILS) {
			if (replyfifo != NULL)
				fifo_calldetails(replyfifo, parameter[0]);
		} else if (s == FIFO_COMMAND_DELETEDETAILS) {
			if (replyfifo != NULL)
				fifo_deletedetails(replyfifo, parameter[0]);
		} else if (s == FIFO_COMMAND_DUMPDEVICES) {
			if (replyfifo != NULL)
				fifo_dumpdevices(replyfifo);
		}		
	}
	
	if (replyfifo != NULL)
		fclose(replyfifo);
	return 0;
}


