/*
 * $Id: IsdnChat.cpp,v 1.5 2003/12/04 10:04:46 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include "globals.h"
#include "IsdnUtils.h"
#include "IsdnChat.h"
#include "IsdnDevice.h"
#include "IsdnDeviceIncoming.h"
#include "IsdnDeviceOutgoing.h"
#include "../../log.h"

/** Creates an empty IsdnChat Object. The state of the instance is the same
 * if reset() is called on a non-empty instance.
 */
IsdnChat::IsdnChat() {
	initialize();
}	

/** Destructor. Frees allocated ressources.
 */
IsdnChat::~IsdnChat() {
	freeList(chatListStart);
	freeList(errorListStart);
}

/** Resets an instance to empty state. This method enables IsdnChat instances 
 * to be reused for a different purpose.
 */
void IsdnChat::reset() {
	freeList(chatListStart);
	freeList(errorListStart);
	initialize();
}	

/** Initializes this IsdnChat instance. It's used by the constructor and 
 * reset(). 
 */
void IsdnChat::initialize() {	
	/* initialize the chat list */
	chatListStart = (chatSequence_t*) malloc (sizeof(chatSequence_t));
	if (chatListStart == NULL) {
		ERROR("out of memory");
		exit(1);
	}	
	chatListStart->next = NULL;
	chatListStart->command = NULL;
	chatListStart->response = NULL;
	chatListStart->options = NO_OPTIONS;

	/* initialize the error handler list */
	errorListStart = (errorList_t*) malloc (sizeof(errorList_t));
	if (chatListStart == NULL) {
		ERROR("out of memory");
		exit(1);
	}	
	errorListStart->next = NULL;
	errorListStart->response = NULL;
	errorListStart->returncode = 0;

	/* set custom timeout defines to "unset" */
	custom_timeout = -1;
	custom_timeout_returncode = FAIL;
}

/** Frees a chat list. Uses recursion to proceed to the end
 * of the list.
 * @param cur chat list to be freed, may not be NULL
 */
void IsdnChat::freeList(chatSequence_t *cur) {
	if (cur->next != NULL) { 
		freeList(cur->next);
	} 
	if (cur->command != NULL) free (cur->command);
	if (cur->response != NULL) free (cur->response);
	free(cur);
}	

/** Frees a error handler list. Uses recursion to proceed to the end
 * of the list.
 * @param cur error list to be freed, may not be NULL
 */
void IsdnChat::freeList(errorList_t *cur) {
	if (cur->next != NULL) { 
		freeList(cur->next);
	} 
	if (cur->response != NULL) free (cur->response);
	free(cur);
}	

/** Sets a custom timeout. If the current chat step could
 * not complete within this timeout, the specified return code
 * is returned from chat().
 * @param tenthsecs the timeout in sec/10 
 * @param returncode the code to be returned on custom timeout
 */
void IsdnChat::setTimeout (int tenthsecs, int returncode) {
	custom_timeout = tenthsecs;
	custom_timeout_returncode = returncode;
}	

/** This chat instance will be used for Initialization. A fail safe chat 
 * sequence is created, which resets the isdn device. \todo maybe a distinction 
 * for different isdn cards is needed here, esp. FCPCI V2.
 */
void IsdnChat::setInitialization() {
	addChatSequence ( "AT", "OK", OPTION_NOT_REQUIRED );       /* clean any garbage */
	addChatSequence ( "ATZ", "OK", OPTION_NOT_REQUIRED );       /* clean any garbage */
	addChatSequence ( "AT", "OK", OPTION_NOT_REQUIRED );       /* clean any garbage */
	addChatSequence ( "AT", "OK" );       /* check if modem emulation responds at all */
	addChatSequence ( "ATH0", "OK", OPTION_NOT_REQUIRED );       /* hangup any existing connection */
	addChatSequence ( "AT&F", "OK" );     /* reset to "factory defaults" */
	addChatSequence ( "ATE0", "OK" );     /* supress local echo */
	addChatSequence ( "AT+FCLASS=8", "OK" );  /* outgoing service type: audio */
	addChatSequence ( "ATS18=1", "OK" );  /* outgoing service type: audio */
	addChatSequence ( "ATS14=4", "OK" );  /* outgoing service type: audio */
	addChatSequence ( "AT+VSM=6", "OK" ); /* select output type for audio */
	addChatSequence ( "ATS13=68", "OK" ); /* extended output and RUNG message */
	addChatSequence ( "ATS23=1", "OK" );  /* extended output including called number on ring */
}	

/** This chat instance will be used for an outgoing call. An ougoing msn 
 * (= number to be displayed as callee on remote pstn side) and a telefone
 * number to be called must be supplied. 
 * @param msn msn to be used
 * @param number to be called -- Note: this number is <B>not</B> checked for
 * validity here.
 */
void IsdnChat::setOutgoingCall (char *msn, char *number) {
	char *dial_command;
	#define DIAL_CMD "ATDT"

	/* set the msn member */
	setMsn(msn);

	assert (number != NULL);

	dial_command = (char *) malloc (strlen(number) + strlen(DIAL_CMD) + 1);
	if (dial_command == NULL) {
		ERROR("out of memory\n");
		exit (1);
	}

	strcpy(dial_command, DIAL_CMD);
	strcat(dial_command, number);
	
	addChatSequence ( dial_command, "" );   /* dial out */
	/* return if connection is established */
	addErrorHandler ( "VCON", DIALING_STATE_CONNECTED );
	/* return if busy signal detected */
	addErrorHandler ( "BUSY", DIALING_STATE_BUSY );
	addErrorHandler ( "NO ANSWER", DIALING_STATE_ERROR );
	addErrorHandler ( "NO CARRIER", DIALING_STATE_ERROR );
	addErrorHandler ( "NO DIALTONE", DIALING_STATE_NODIALTONE );
	/* try a maximum of three secs, if no busy until now, we assume ringing 
	 * \todo maybe there is a better way to determine ringing state
	 */
	setTimeout(5, DIALING_STATE_RINGING);
		
// addChatSequence ( "AT+VTX+VRX", "CONNECT", OPTION_NOT_REQUIRED );   /* enable voice transmission */
	free(dial_command);
}	

/** Connect outoing audio. If a isdn connection is established, try to setup
 * the duplex audio connection.
 */
void IsdnChat::setConnect () {
	/* \todo extensive tests with Fritz V2.0, does not reply 
	 * to AT+VRX+VTX, by now OPTION_NOT_REQUIRED is used. */
	addChatSequence ( "AT+VRX+VTX", "CONNECT", OPTION_NOT_REQUIRED );   /* enable voice transmission */
	addErrorHandler ( "NO ANSWER", DIALING_STATE_ERROR );
	addErrorHandler ( "NO CARRIER", DIALING_STATE_ERROR );
	/* this must work within 1 second */
	setTimeout ( 10, FAIL );
}

/** Answer incoming call. 
 */
void IsdnChat::setAnswerIncomingCall () {
	addChatSequence ( "ATA", "CONNECT", OPTION_NOT_REQUIRED );   /* enable voice transmission */
	/* this must work within 1 second */
	setTimeout ( 10, FAIL );
}

/** Pseudo chat. Useful in combination with addErrorHandler() and
 * setTimeout().
 */
void IsdnChat::setWaitForAnswer () {
	addChatSequence ( "", "" );
}

/** Set ougoing msn chat item. 
 * @param msn the msn to be used for outgoing calls.
 */
void IsdnChat::setMsn (char *msn, bool listen) {
	char *msn_command;
	#define MSN_CMD "AT&E"
	#define MSN_LISTEN_CMD "AT&L"
	assert (msn != NULL);
	msn_command = (char *) malloc (strlen(msn) + strlen(MSN_CMD) + 1);
	if (msn_command == NULL) {
		ERROR("out of memory\n");
		exit (1);
	}
	if (listen)
		strcpy(msn_command, MSN_LISTEN_CMD);	
	else
		strcpy(msn_command, MSN_CMD);	
	strcat(msn_command, msn);
	addChatSequence ( msn_command, "OK" );    /* set msn */
	free(msn_command);
}

/** End voice and hangup chat item. Failsafe implementation that
 * ends voice connection if open and then hangs up connection if
 * open.
 */ 
void IsdnChat::setHangup() {
	addChatSequence("","", OPTION_END_VOICE | OPTION_NOT_REQUIRED );
	addChatSequence("","", OPTION_END_VOICE | OPTION_NOT_REQUIRED );
	addChatSequence("ATH0","OK", OPTION_NOT_REQUIRED);
	addChatSequence("ATH0","OK", OPTION_NOT_REQUIRED);
}	

/** Set device to listen for incoming calls on specified number. This is 
 * not implemented yet.
 * @param msn number to listen for calls on
 */
void IsdnChat::setIncomingDevice (char *msn) {
	setMsn(msn, true);
}	

/** Append a chat item to the list. Can be used directly or indirect by setter
 * methods.
 * @param command the command to be sent to the isdn device, or an empty 
 * string to not send anything.
 * @param response the response to be waited for, empty for not waiting for 
 * an answer.
 * @param options special options for this item, e.g. OPTION_NOT_REQUIRED or 
 * OPTION_END_VOICE
 */
void IsdnChat::addChatSequence (char* command, char* response, char options) {
	chatSequence_t *temp;
	temp = chatListStart;

	if (temp->command != NULL) {
		while (temp->next != NULL) {   /* find end of list */
			temp = temp->next;
		}	
		/* create new entry */
		temp->next = (chatSequence_t *) malloc (sizeof (chatSequence_t));
		if (temp->next == NULL) {
			ERROR("out of memory\n");
			exit(1);
		}	
		temp = temp->next;
	}	

	/* this is last in list */
	temp->next = NULL;
	/* set properties */
	temp->command = (char*) malloc (strlen(command) + 1);
	temp->response = (char*) malloc (strlen(response) + 1);
	if ((temp->command == NULL) || (temp->response == NULL)) {
		ERROR("out of memory\n");
		exit(1);
	}	
	strcpy(temp->command, command);
	strcpy(temp->response, response);
	temp->options = options;
}

/** Add a custom error handler. The specified response is searched for
 * in every answer, if it occurs, chat returns immediately with the specified 
 * returncode.
 * @param response string to be watched for
 * @param returncode returncode of chat() if response is found
 */
void IsdnChat::addErrorHandler (char* response, int returncode) {
	errorList_t *temp;
	temp = errorListStart;

	if (temp->response != NULL) {
		while (temp->next != NULL) {   /* find end of list */
			temp = temp->next;
		}	
		/* create new entry */
		temp->next = (errorList_t *) malloc (sizeof (errorList_t));
		if (temp->next == NULL) {
			ERROR("out of memory\n");
			exit(1);
		}	
		temp = temp->next;
		
	}
	
	/* this is last in list */
	temp->next = NULL;
	/* set properties */
	temp->response = (char*) malloc (strlen(response) + 1);
	if (temp->response == NULL) {
		ERROR("out of memory\n");
		exit(1);
	}	
	strcpy(temp->response, response);
	temp->returncode = returncode;
}

/** Check if any custom error handler matches. 
 * @param response the string to be compared with all error strings
 */
int IsdnChat::check(char* response) {
	errorList_t* cur;
	cur = errorListStart;
	
	if (cur->response == NULL) 
		return (OK);

	/* go through list, return specified code on match */
	while (cur != NULL) {
		if (strcmp(response, cur->response) == 0) {
			return (cur->returncode);
		}
		cur = cur->next;
	} /* while */
	return (OK);
}	

/** Dumps the list to debugging output. Normally unused debugging method.
 */
void IsdnChat::debugList() {
	chatSequence_t *temp;
	int i;
	i = 0;
	temp = chatListStart;

	while (temp != NULL) {
		INFO("ITEM #%i@%p: %s -> %s options: %i next@%p\n", i++, temp, temp->command, temp->response, temp->options, temp->next);
		temp = temp->next;
	}	
}	

/** Chat with isdn device. The actual chat conversation specified by
 * one ore more of the other methods is run. Takes care of custom timeouts and
 * error handlers.
 * @param isdn_fd file handle of open isdn device to be used
 * @return OK if the sequence completes normally <BR>
 * @return FAIL if a generic error occurs
 * @return other_value if custom error handler or custom timeout hits
 */
int IsdnChat::chat (int isdn_fd) {
	char buf[BUFSIZE];
	chatSequence_t *current;
	int ret;
	
	current = chatListStart;

	/* process all steps in list */
	while (1) {
		int retry = 0;

		/* construct command to write out */
		int doWrite = OK;
		
		/* deal with special options first */
		#define END_VOICE_SEQUENCE "\x010\x014\x010\x003"
		if (current->options & OPTION_END_VOICE) {  /* end voice connection, special string, no lf */
			strcpy(buf, END_VOICE_SEQUENCE);
		} else {
			/* no special options: append a linefeed to current command */
			assert(strlen(current->command) < BUFSIZE);
			strcpy(buf, current->command);
			strcat(buf, "\n");
			if (strlen (current->command) <= 0 ) 
				doWrite = FAIL;
		}	

		/* do nonblocking write if command is not zero-sized length */
		if ( (doWrite == OK) && ( IsdnUtils::nb_write(isdn_fd, buf) == FAIL )) {
			return FAIL;
		}	

		/* nonblocking-read the last answer from modem */
		ret = IsdnUtils::nb_read_answer(isdn_fd, buf, current->options, custom_timeout);
		if ( ( ret == FAIL) || (ret == -2 ) ) {
			/* read failed */
			if (current->options & OPTION_NOT_REQUIRED) {
				//DBG("no answer for >>%s<<. Ignoring.\n", current->command);
				goto next;
			}	
				
			if ( ret == -2 ) { /* timeout */
				//DBG("custom timeout hit while >>%s<<, exiting\n", current->command);
				return custom_timeout_returncode;
			}	
			#define LIOPTION_RETRY_LIMIT 3
			if ( (current->options & OPTION_RETRY) && (retry++ < LIOPTION_RETRY_LIMIT) ) {
				INFO("retrying to read an answer from modem for >>%s<<\n", current->command);
				continue;
			}	
			INFO("command failed >>%s<<\n", current->command);
			return (FAIL);
		}	else {
			/* read ok */

			/* check for "expected" failure response */
			int f;
			if ((f = check(buf)) != OK)
				return (f); /* matched errorHandler -> return specified code */

			if (current->options & OPTION_END_VOICE) {
				// don't check result
				goto next;
			}	
			
			/* check if answer matches the expected one */	
			if (strcmp(buf, current->response) == 0) {
				/* yap, fine. */
//				DBG("%s -> %s\n", current->command, buf);
				goto next;
			} else {
				/* no match */
				if (current->options & OPTION_NOT_REQUIRED) {
					DBG("ignoring wrong answer >>%s<< to optional command >>%s<<\n", buf, current->command);
					goto next;
				} else {	
					ERROR("received wrong answer >>%s<< to required command >>%s<<\n", buf, current->command);
					return (FAIL);
				}	
			}	
		} /* if read */
		/* we may never get here */
		assert(false);
		
		next: 
			/* process next chat pair, if exists */
			if (current->next != NULL) {
				current = current->next;
			continue;
			} else {
				/* thats it */
			break;
		}	
	} /* while */
	/* everythings fine */
	return (OK);
} /* chat() */


