/*
 * $Id: IsdnConfiguration.cpp,v 1.11 2004/04/05 12:44:50 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "SemsConfiguration.h"
#include "globals.h"
#include "log.h"
#include "AmApi.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>

/** Globally available instance of IsdnConfiguration */
IsdnConfiguration* isdn_config;
		
/** Constructor. The configuration is not read here yet.
 */
IsdnConfiguration::IsdnConfiguration() {
	int i;
	i4l_outdevices = NULL; /* set initially to NULL to make realloc work */
	i4l_indevices = NULL; /* set initially to NULL to make realloc work */
	out_msn = NULL; 
	out_logfile = NULL;
	firstfreedevice = 0;
	minnumberlen = 4;

	for (i=0; i<MAX_CAPI_CONTROLLERS; i++) {
		devices[i].inboundPatterns = NULL;
		devices[i].outboundPatterns = NULL;
		devices[i].numberOfBChannels = -1;
		devices[i].available = FAIL;
		devices[i].availBChannels = -1;
		*(devices[i].identification) = '\0';
	}

	strcpy(devices[0].identification, "ISDN4LINUX");
}

IsdnConfiguration::~IsdnConfiguration() {
	#ifdef USE_CAPI20
	delete(capiManager);
	#endif
	remove_lockfiles(i4l_indevices);
	remove_lockfiles(i4l_outdevices);
}

/** Make shure the globally configuration is a valid instance of 
 * IsdnConfiguration.
 * This method does nothing in 2nd or subsequent calls
 */
void IsdnConfiguration::instantiate() {
	if (!isdn_config) {
		isdn_config = new IsdnConfiguration();
	} else {
		assert(!isdn_config);
	}	
}	

void IsdnConfiguration::reload() {
	#ifndef USE_CAPI20
	char* c = " (supported backends: isdn4linux)";
	#else
	char* c = " (supported backends: CAPI20 isdn4linux)";
	#endif
	INFO("isdngw version %s loading%s ...\n", ISDNGW_VERSION, c);
	
	/* initialize optional parameters to defaults */
	lockdir = (char *) malloc (strlen(LOCKDIR_DEFAULT) + 1);
	if (lockdir == NULL) {
		ERROR("out of memory\n");
		exit(1);
	}
	strcpy(lockdir, LOCKDIR_DEFAULT);
	if (reloadModuleConfig(MODULE_NAME) == 0) {
		exit(-1);
	}	
	warnUnknownParams();
}

/** Removes all lock files.
 */
void IsdnConfiguration::remove_lockfiles (struct list_t* dlist) {
	if (!dlist) {
		return;
	}	
	remove_lockfiles(dlist->next);
	((IsdnDevice*)(dlist->link))->removeLockFile();
}	

/** Creates a device list consisting of num devices starting at /dev/ttyI<start>.
 */
struct list_t* IsdnConfiguration::createI4lDeviceList (int num, int start) {
	struct list_t* list;
	struct list_t* listpos;
	int i;
	int len;

	if (num == 0) /* no devices */
		return NULL;

	len = strlen (TTYI_DEVICE_PREFIX) + 3; /* prefix + number + \0 */

	list = (struct list_t*) malloc (sizeof (struct list_t*));
	if (list == NULL) {
		ERROR("no memory for device list\n");
		return NULL;
	}	
	listpos = list;

	for (i=start; i<start+num; i++) {
		listpos->item = (char*) malloc (len);
		sprintf(listpos->item, "%s%i", TTYI_DEVICE_PREFIX, i);
		listpos->link = NULL;
		listpos->next = NULL;
		if (i < start+num-1) {
			listpos->next = (struct list_t*) malloc (sizeof (struct list_t));
			if (listpos->next == NULL) {
				ERROR("no memory for next list item\n");
				return NULL;
			}	
			listpos=listpos->next;
		}	
	}	
	return list;
}

/* creates a list from a string like "item1, item2,item3   ,  item4" */
struct list_t* IsdnConfiguration::createList (char* list) {
	struct list_t* retval;
	retval = (struct list_t*) malloc (sizeof (struct list_t));
	if (retval == NULL) {
		ERROR("no memory to create list\n");
		return NULL;
	}
	if (!_createList(retval, list)) {
		return NULL;
	}
	return retval;
}	

/* free a list */
void IsdnConfiguration::freeList (struct list_t* list) {
	if (list->next == NULL) {
		/* end of list */
		free(list);
	} else {
		freeList(list->next);
		free(list);
	}
}	

/** Creates a char* list of a comma seperated list using recursion.
 * @param node start of the list (must be allocated already)
 * @param list comma seperated list of numbers ("814123123,123123814")
 * @return true on success, false on ",," in the list
 */
bool IsdnConfiguration::_createList (struct list_t* node, char* list) {
	char* local_pointer = list;
	char c;

	assert (node != NULL);
	while (1) {
		c = *(local_pointer++);
		if ( c == ',' || c == '\0') {
			*(local_pointer-1) = '\0';   /* terminate the current number "list" */
			break;
		}	
	}
	
	/* check for zero sized number */
	if (strlen(list) == 0) 
		return false;
	
	/* store the item in the structure */
	node->item = (char*) malloc (strlen(list)+1);
	if (node->item == NULL) { ERROR("out of memory"); exit (1); }
	strcpy(node->item, list);
	node->link = NULL;

	/* more devices: go into recursion with rest of string */
	if (c == ',') {
		node->next = (struct list_t*) malloc (sizeof (struct list_t));
		if (node->next == NULL) { ERROR("out of memory"); exit (1); }
		return _createList(node->next, local_pointer);
	} else {
		node->next = NULL;
	}	
	return true;
}	
	
int IsdnConfiguration::check() {
	char *temp;
	int i;
	int num;

	/* required parameters */
	bool enablei4l_set = false;
	bool enablecapi_set = false;
	bool callerdomain_set = false;
	bool calleedomain_set = false;
	
	/* only one of two allowed parameters */
	bool out_dev_set = false;
	bool numout_dev_set = false;
	bool in_dev_set = false;
	bool numin_dev_set = false;

	clip = true;

	
	/*************************/
	/* backend selection     */
	/*************************/

	/* enablei4l */
	temp = getValueForKey("enablei4l");
	if (temp != NULL) {
		if (strcasecmp(temp, "yes") == 0)
			enablei4l = true;
		else if (strcasecmp(temp, "no") == 0)
			enablei4l = false;
		else
			reportConfigError("enablei4l", "please specify exactly yes or no", true);
		enablei4l_set = true;
	}		
	
	/* enablecapi */
	temp = getValueForKey("enablecapi");
	if (temp != NULL) {
		if (strcasecmp(temp, "yes") == 0) {
			#ifdef USE_CAPI20
			enablecapi = true;
			#else
			reportConfigError("enablecapi", "CAPI support disabled at compile time, rebuild with -DUSE_CAPI20 first.", true);
			#endif
		} else if (strcasecmp(temp, "no") == 0) {
			enablecapi = false;
		} else {
			reportConfigError("enablei4l", "please specify exactly yes or no", true);
		}	
		enablecapi_set = true;
	}		
	
	if ((!enablecapi) && (!enablei4l))
		reportConfigError((char*)NULL, "no backend selected, you need at least one backend to run isdngw (enablecapi=yes or enablei4l=yes)", true);	
	
	/*************************/
	/* isdn4linux parameters */
	/*************************/
	if (enablei4l) {
	
		/* i4l_numoutdevices */
		temp = getValueForKey("i4l_numoutdevices");
		if ((temp != NULL) && (strlen(temp) != 0)) {
			if (out_dev_set)
				reportConfigError("i4l_numoutdevices", "you may only specify i4l_outdevices OR i4l_numoutdevices, not both", true);
			if (sscanf(temp, "%i", &num) != 1) 
				reportConfigError("i4l_numoutdevices", "could not parse number", true);
			i4l_outdevices = createI4lDeviceList(num, firstfreedevice);
			firstfreedevice+=num;
			if (i4l_outdevices == NULL)
				reportConfigError("i4l_outdevices", "internal error", true);
			numout_dev_set = true;
		}	

		/* i4l_outdevices */
		temp = getValueForKey("i4l_outdevices");
		if ((temp != NULL) && (strlen(temp) != 0)) {
			if (numout_dev_set)
				reportConfigError("i4l_outdevices", "you may only specify i4l_outdevices OR i4l_numoutdevices, not both", true);
			i4l_outdevices = createList(temp);
			if (i4l_outdevices == NULL)
				reportConfigError("i4l_outdevices", "empty device name specified or other parser error", true);
			out_dev_set = true;
		}	

		/* i4l_numindevices */
		temp = getValueForKey("i4l_numindevices");
		if ((temp != NULL) && (strlen(temp) != 0)) {
			if (in_dev_set)
				reportConfigError("i4l_numindevices", "you may only specify i4l_indevices OR i4l_numindevices, not both", true);
			if (sscanf(temp, "%i", &num) != 1) 
				reportConfigError("i4l_numindevices", "could not parse %s as number", true);
			i4l_indevices = createI4lDeviceList(num, firstfreedevice);
			firstfreedevice+=num;
			if (i4l_indevices == NULL)
				reportConfigError("i4l_indevices", "internal error", true);
			in_dev_set = true;
		}	

		/* i4l_indevices */
		temp = getValueForKey("i4l_indevices");
		if ((temp != NULL) && (strlen(temp) != 0)) {
			if (numin_dev_set)
				reportConfigError("i4l_indevices", "you may only specify i4l_indevices OR i4l_numindevices, not both", true);
			i4l_indevices = createList(temp);
			if (i4l_indevices == NULL)
				reportConfigError("i4l_indevices", "empty device name specified or other parser error", true);
			in_dev_set = true;
		}	

		/* i4l_lockdir */
		temp = getValueForKey("i4l_lockdir");
		if (temp != NULL)
			lockdir = temp;

		/* i4l_inmsn */
		temp = getValueForKey("i4l_inmsn");
		if (temp != NULL) {
			if ((temp != NULL) && (setDeviceInboundMapping(DEVICE_I4L, temp) == FAIL))
				reportConfigError("i4l_inmsn", errorAsString);
		}

		/* i4l_mapping */
		temp = getValueForKey("i4l_mapping");
		i = setDeviceOutboundMapping(DEVICE_I4L, temp);
		if (i == FAIL)
			reportConfigError("i4l_mapping", errorAsString, true);

		/* i4l_numchannel */
		temp = getValueForKey("i4l_numchannel");
		if (setNumberOfBChannels(DEVICE_I4L, temp) == FAIL)
			reportConfigError("i4l_numchannel", errorAsString, true);

	} else {
	/* check for all parameters, needed to avoid warnings about unknown parameters */
		temp = getValueForKey("i4l_numoutdevices");
		temp = getValueForKey("i4l_outdevices");
		temp = getValueForKey("i4l_numindevices");
		temp = getValueForKey("i4l_indevices");
		temp = getValueForKey("i4l_lockdir");
		temp = getValueForKey("i4l_inmsn");
		temp = getValueForKey("i4l_mapping");
		temp = getValueForKey("i4l_numchannel");
	}
	
	/*************************/
	/* capi parameters       */
	/*************************/
	
	if (enablecapi) {

		/* capi_*_inmsn */
		for (i=1; i<=MAX_CAPI_CONTROLLERS; i++) {
			char paramname[30];
			sprintf(paramname, "capi_%i_inmsn", i);
			temp = getValueForKey(paramname);
			if ((temp != NULL) && (setDeviceInboundMapping(i, temp) == FAIL))
				reportConfigError(paramname, errorAsString);
		}

		/* capi_*_mapping */
		for (i=1; i<=MAX_CAPI_CONTROLLERS; i++) {
			char paramname[30];
			sprintf(paramname, "capi_%i_mapping", i);
			temp = getValueForKey(paramname);
			if ((temp != NULL) && (setDeviceOutboundMapping(i, temp) == FAIL))
				reportConfigError(paramname, errorAsString);
		}		

		/* capi_*_numchannel */
		for (i=1; i<=MAX_CAPI_CONTROLLERS; i++) {
			char paramname[30];
			sprintf(paramname, "capi_%i_numchannel", i);
			temp = getValueForKey(paramname);
			if ((temp != NULL) && (setNumberOfBChannels(i, temp) == FAIL))
				reportConfigError(paramname, errorAsString);
		}

	} else {
	/* check for all parameters, needed to avoid warnings about unknown parameters */
		for (i=1; i<=MAX_CAPI_CONTROLLERS; i++) {
			char paramname[30];
			sprintf(paramname, "capi_%i_inmsn", i);
			temp = getValueForKey(paramname);
			sprintf(paramname, "capi_%i_mapping", i);
			temp = getValueForKey(paramname);
			sprintf(paramname, "capi_%i_numchannel", i);
			temp = getValueForKey(paramname);
		}
	}	

	/*************************/
	/* SIP parameters        */
	/*************************/

	/* msnsipcaller */
	temp = getValueForKey("msnsipcaller");
	if (temp != NULL) {
		if (strcasecmp(temp, "yes") == 0)
			msnsipcaller = true;
		else if (strcasecmp(temp, "no") == 0)
			msnsipcaller = false;
		else
			reportConfigError("msnsipcaller", "please specify exactly yes or no", true);
	}
	
	/* defaultmsn */
	temp = getValueForKey("defaultmsn");
	if (temp != NULL) {
		out_msn = temp;
	}	

	/* callerdomain */
	temp = getValueForKey("callerdomain");
	if (temp != NULL) {
		callerdomain = temp;
		callerdomain_set = true;
	}	

	/* calleedomain */
	temp = getValueForKey("calleedomain");
	if (temp != NULL) {
		calleedomain = temp;
		calleedomain_set = true;
	}	

	/* clip */
	temp = getValueForKey("clip");
	if ((temp != NULL) && (strlen(temp) != 0)) {
		if (strcasecmp(temp, "yes") == 0)
			clip = true;
		else if (strcasecmp(temp, "no") == 0)
			clip = false;
		else
			reportConfigError("clip", "please specify exactly yes or no", true);
	}
		
	/* minnumberlen */
	temp = getValueForKey("minnumberlen");
	if ((temp != NULL) && (strlen(temp) != 0)) {
		if (sscanf(temp, "%i", &minnumberlen) != 1) {
			reportConfigError("minnumberlen", "could not parse number");
		}		
	}	
		
	/* outlogfile */
	temp = getValueForKey("outlogfile");
	if (temp != NULL)
		out_logfile = temp;
	
	if (!enablei4l_set) reportConfigError((char*)NULL, "required parameter enablei4l missing", true);
	if (!enablecapi_set) reportConfigError((char*)NULL, "required parameter enablecapi missing", true);
	if (!callerdomain_set) reportConfigError((char*)NULL, "required parameter callerdomain missing", true);
	if (!calleedomain_set) reportConfigError((char*)NULL, "required parameter calleedomain missing", true);

/* extra debugging for device selection */
/*
	DBG("dumping outdevs\n");
	debugDumpList(i4l_outdevices);
	DBG("dumping indevs\n");
	debugDumpList(i4l_indevices);
	DBG("dumping sortedInBound\n");
	debugDumpList(sortedInboundPatterns);
	DBG("dumping sortedOutBound\n");
	debugDumpList(sortedOutboundPatterns);
*/

	return 1;
}	

int IsdnConfiguration::setDeviceOutboundMapping(int device, char* list) {
	return _setDeviceMapping(device, list, false);
}

int IsdnConfiguration::setDeviceInboundMapping(int device, char* list) {
	return _setDeviceMapping(device, list, true);
}

int IsdnConfiguration::_setDeviceMapping(int device, char* list, bool inbound) {
	struct list_t* l;
	if (strlen(list) == 0)
		return OK;
	l = createList(list);
	if (l == NULL) {
		strcpy(errorAsString, "parse error");
		return FAIL;
	}
	devices[device].available = OK;
	if (inbound) {
		devices[device].inboundPatterns = l;
	} else {
		devices[device].outboundPatterns = l;
	}	
	addNumberPatterns(device, l, inbound);
	return OK;
}

/* insert the numbers from the list in the right sorted pattern list */
void IsdnConfiguration::addNumberPatterns(int device, struct list_t* list, bool inbound) {
	struct list_t* plist;
	struct list_t* pos;
	
	/* set plist to the right list, create the list first if needed */
	if (inbound) {
		if (sortedInboundPatterns == NULL) {
			sortedInboundPatterns = (struct list_t*) malloc (sizeof(struct list_t));
			if (sortedInboundPatterns == NULL) {
				ERROR("no memory to create sortedInboundPatterns\n");
				return;
			}	
			sortedInboundPatterns->item = NULL;
			sortedInboundPatterns->next = NULL;
			sortedInboundPatterns->link = NULL;
		}	
		plist = sortedInboundPatterns;
	} else {			
		if (sortedOutboundPatterns == NULL) {
			sortedOutboundPatterns = (struct list_t*) malloc (sizeof(struct list_t));
			if (sortedOutboundPatterns == NULL) {
				ERROR("no memory to create sortedOutboundPatterns\n");
				return;
			}	
			sortedOutboundPatterns->item = NULL;
			sortedOutboundPatterns->next = NULL;
			sortedOutboundPatterns->link = NULL;
		}	
		plist = sortedOutboundPatterns;
	}	
	
	pos = list;
	do {
		addSorted(plist, pos, device);
		pos = pos->next;
	} while (pos != NULL);
}

/* add cur into target in a sorted manner. sorting is done by item-strlen */
void IsdnConfiguration::addSorted(struct list_t* target, struct list_t* cur, int device) {
	struct list_t* pos;
	struct list_t* newnode;
	unsigned curlen;

	curlen = strlen(cur->item);
	pos = target;
	while (1) {
		if (pos->item == NULL) { /* emtpy list */
			pos->item = cur->item;
			pos->num = device;
			break;
		} else {
			if (strlen(pos->item) < curlen) { /* insert before pos */
				newnode = (struct list_t*) malloc (sizeof (struct list_t));
				if (newnode == NULL) {
					ERROR("no memory for newnode\n");
					return;
				}	
				newnode->next = pos->next;
				newnode->item = pos->item;
				newnode->num = pos->num;
				pos->next = newnode;
				pos->item = cur->item;
				pos->num = device;
				break;
			} else { /* go to next node */
				if (pos->next != NULL) {
					pos = pos->next;
					continue;
				} else { /* add a new node for the current entry */
					newnode = (struct list_t*) malloc (sizeof (struct list_t));
					if (newnode == NULL) {
						ERROR("no memory for newnode\n");
						return;
					}	
					pos->next = newnode;
					newnode->item = NULL;
					newnode->link = NULL;
					newnode->next = NULL;
					pos = newnode;
					continue;
				}
			}	
		}
	}	
}

int IsdnConfiguration::setNumberOfBChannels(int device, char* number) {
	int num;
	if (strlen(number) == 0)
		num = -1;
	else if (sscanf(number, "%i", &num) != 1) {
		strcpy(errorAsString, "could not parse number");
		return FAIL;
	}
	devices[device].available = OK;
	devices[device].numberOfBChannels = num;
	devices[device].availBChannels = num;
	return OK;
}

void IsdnConfiguration::debugDumpList (struct list_t* list) {
	int i = 0;
	do {
		DBG("LIST: item# %i num=%i item=\"%s\" link@%p\n", i++, list->num, list->item, list->link);
		list = list->next;
	} while (list != NULL);
}	

void IsdnConfiguration::getListAsString (struct list_t* list, char* buf, int len) {
	int offset = 0;
	if (list == NULL) {
		*buf = '\0';
		return;
	}	
	do {
		int l = strlen(list->item);
		if (l + offset + 1 < len) {
			sprintf(buf+offset, "%s", list->item);
			offset+=l;
		}	else {
			ERROR("output buffer to small, debug string truncated\n");
			return;
		}	
		list = list->next;
		if (list != NULL) {
			sprintf(buf+offset, " ");
			offset+=1;
		}	
			
	} while (list != NULL);
}	

void IsdnConfiguration::debugDumpDeviceList () {
	int i;
	for (i=0; i<=MAX_CAPI_CONTROLLERS; i++) {
		DBG("################################ DUMPING CONTROLLER %i\n", i);
		DBG("available=%s numberOfBChannels=%i availBChannels=%i\n", devices[i].available==OK?"yes":"no ", devices[i].numberOfBChannels, devices[i].availBChannels);
		DBG("inboundPatterns:\n");
		debugDumpList(devices[i].inboundPatterns);
		DBG("outboundPatterns:\n");
		debugDumpList(devices[i].outboundPatterns);
	}
}

void IsdnConfiguration::dumpConfiguredDevices () {
	int i;
	char buf1[1000];
	char buf2[1000];
	
	for (i=0; i<=MAX_CAPI_CONTROLLERS; i++) {
		if (devices[i].available==OK) {
			getListAsString(devices[i].inboundPatterns, buf1, 1000);	
			getListAsString(devices[i].outboundPatterns, buf2, 1000);	
			DBG("%s dev %i up listening on [%s], calling to [%s].\n", i==DEVICE_I4L?"i4l ":"capi", i, buf1, buf2);
		}
	}
}

void IsdnConfiguration::fifo_dumpdevices(FILE *reply) {
	int i;
	for (i=0; i<=MAX_CAPI_CONTROLLERS; i++) {
		if (devices[i].available==OK) {
			if (devices[i].availBChannels != -1)
				fprintf(reply, "%i: %s device (%s, %i b-channels)\n", i, i==DEVICE_I4L?"I4L":"CAPI", devices[i].identification, devices[i].availBChannels);
			else
				fprintf(reply, "%i: %s device (%s, unknown number of b-channels)\n", i, i==DEVICE_I4L?"I4L":"CAPI", devices[i].identification);

		}
	}
}	

int IsdnConfiguration::getDeviceForCall(char* tonumber, char* fromnumber) {
//	DBG("checking which device to use for call from %s to %s\n", fromnumber, tonumber);
	return getLongestMatch(sortedOutboundPatterns, tonumber);
}	

int IsdnConfiguration::acceptCallOnDevice(int device, char* number) {
	char* cur;
	int len;
	int finished = -1;
	int match = -1;
	unsigned int complen;
	struct list_t* pos;
	DBG("checking whether to accept the call to %s on device %i ...\n", number, device);
	pos = devices[device].inboundPatterns;
	while ((pos != NULL) && (pos->item != NULL)) {
		cur = pos->item;
		len = strlen(cur);
		if (len < finished) /* done */
			return match;
		
		if ( cur[len-1] == '*' ) { /* check if ends in "*" */
			complen = len - 1;
		} else {
			complen = len;
			if (complen < strlen(number)) { /* cannot match! */
				pos = pos->next;
				continue;
			}	
		}	
		if (strncmp(cur, number, complen) == 0) {
			return OK;
		}
		pos = pos->next;
	}	
	return FAIL;
}	


int IsdnConfiguration::getLongestMatch(struct list_t* list, char* number) {
	list_t* pos;
	char* cur;
	int len;
	int finished = -1;
	int match = -1;
	unsigned int complen;
	
	pos = list;
	while ((pos != NULL) && (pos->item != NULL)) {
		cur = pos->item;
		len = strlen(cur);
		if (len < finished) /* done */
			return match;
		
		if ( cur[len-1] == '*' ) { /* check if ends in "*" */
			complen = len - 1;
		} else {
			complen = len;
			if (complen < strlen(number)) { /* cannot match! */
				pos = pos->next;
				continue;
			}	
		}	
		if (strncmp(cur, number, complen) == 0) {
		//	DBG("comparison between %s and %s matches for complen=%i chars\n", cur, number, complen);
			if (devices[pos->num].available == FAIL) {
				ERROR("configured device %i would be responsible for this call, but it is unavailable (ignoring)\n", pos->num);
			} else {	
				DBG("device %i responsible for this call, because %s matches%s\n", pos->num, cur, match!=-1?" (ignoring, already have a device)":"");
				if (match == -1)
					match = pos->num;
				finished = complen;
			}
		}
		pos = pos->next;
	}	
	return match;
}
