/*
 * $Id: IsdnDevice.cpp,v 1.5 2004/04/05 12:08:37 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "IsdnDevice.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "globals.h"
#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"
#include "log.h"
#include "AmSession.h"

/** Create a new IsdnDevice for incoming or outgoing connections. 
 * Sets up synchronization and some other prerequisits, the device is 
 * not opened yet.
 * @param devicename file name of the isdn device, e.g. /dev/ttyI42
 */
IsdnDevice::IsdnDevice(char* devicename) {
	/* allocate and set the devicename */
	m_devicename = (char *) malloc (strlen(devicename)+1);
	if (m_devicename == NULL) {
		ERROR("out of memory\n");
		exit(1);
	}	
	strcpy(m_devicename, devicename);
	
	/* construct lockfilename, take text after last slash for postfix */
	char* temp = m_devicename;
	int lastslash = -1; /* not found */
	while ( *(temp++) != '\0' ) {
		if ( *temp == '/' )
			lastslash = temp - m_devicename;
	}
	if (lastslash == -1) {
		ERROR("could not determine lockfile from devicename %s\n", m_devicename);
		m_lockfilename = NULL;
	}	

	int lockfile_len = strlen (isdn_config->lockdir) 
					+ strlen(LOCKFILE_PREFIX) 
					+ strlen(m_devicename + lastslash) + 1;
	m_lockfilename = (char *) malloc (lockfile_len);
	if (m_lockfilename == NULL) {
		ERROR("out of memory\n");
		exit(1);
	}
	/* "/var/lock" + "/LCK.." + "ttyI10" */
	strcpy(m_lockfilename, isdn_config->lockdir);
	strcat(m_lockfilename, LOCKFILE_PREFIX);
	strcat(m_lockfilename, m_devicename+lastslash+1);

	m_status = STATUS_UNINIT;

	m_request_hangup = FAIL;

	pthread_mutex_init(&mut_dev, NULL);
	pthread_mutex_init(&mut_hup, NULL);

}

/** Destructor. Closes the device, removes the lockfile and frees
 * all allocated ressources.
 */
IsdnDevice::~IsdnDevice() {
	close(m_fd);
	removeLockFile();
	DBG("device %s destroyed\n", m_devicename);	
	if (m_devicename != NULL)
		free(m_devicename);
	if (m_lockfilename != NULL)
		free(m_lockfilename);
//	pthread_mutex_unlock(&mut_dev);
	pthread_mutex_destroy(&mut_dev);
	pthread_mutex_destroy(&mut_hup);

	/* subclass cleanup */
	cleanup();
}


void IsdnDevice::setCall(IsdnCall* call) {
	m_call = call;
}	

IsdnCall* IsdnDevice::getCall() {
	return m_call;
}	

/** Try to lock the device. If the device is unsused it is locked and
 * ready for use.
 * @return OK now the device is locked and ready to be used <BR>
 * @return FAIL the device is already locked, it cannot be used
 */
int IsdnDevice::trylock() {
	int s;
	s = pthread_mutex_trylock(&mut_dev);
	if (s == EBUSY) {
		return (FAIL);
	}	
	return (OK);
}

/** (Re-)initialize a device. A device must be initialized before use. A call 
 * to this method must always work, no matter in what state the device
 * currently is.
 * @return OK on success <BR>
 * @return FAIL on error 
 */ 
int IsdnDevice::initialize() {

	/* create lock file if not locked already */
	if ( ! (m_status & STATUS_LOCKED) ) {
		if (createLockFile() == FAIL) {
			INFO("(re-)initializer: lock file error (%i & %i -> %i)\n", m_status, STATUS_LOCKED,  m_status & STATUS_LOCKED);
			return (FAIL);
		}	
	}	

	/* open device for nonblocking i/o if not open already */
	if ( ! (m_status & STATUS_OPEN) ) {
		if ( open_device() == FAIL) {
			INFO("(re-)initializer: file error\n");
			return (FAIL);
		}	
	}

	/* set terminal parameters, esp. "raw mode" */
	if (setTerminalParameters() == FAIL) {
		INFO("(re-)initializer: termios error\n");
		return (FAIL);
	}	
	
	/* make shure the buffers are empty */
	flush();

	/* TODO: check if this is enough */
	
	resetHangupRequested();
	
	/* initialize device */
	IsdnChat ic = IsdnChat();
	ic.setInitialization();
	if (ic.chat(m_fd) == FAIL) {
		if (ic.chat(m_fd) == FAIL) {
			if (ic.chat(m_fd) == FAIL) {
				ERROR("initialization failed - terminating\n");
				exit(-1);
			}
		}
	} else {
		DBG("device %s (%s) initialized\n", m_devicename, getType());
	}
	return (OK);
}

/** Immediately end a connection. Voice transfer is stopped and device is 
 * hung up. This must always work, no matter in what state a device is. If no
 * connection is up, this method does nothing, but flushing the device.
 */
void IsdnDevice::hangup() {
	/* on call termination we get here, either pstn hangup or SIP:bye */
	IsdnChat ic = IsdnChat();
	int ret;
	flush();
	ic.setHangup();
	if ((ret = ic.chat(m_fd)) == FAIL) {
		DBG("hangup failed\n");
	}	
}

/** Unlock a previously locked device. No operations may be performed
 * on nonlocked devices.
 */
void IsdnDevice::unlock() {
	pthread_mutex_unlock(&mut_dev);
}

/** Tell the device to hangup the pending call. On the next call to getClear()
 * OK will be returned, meaning "end the call".
 */ 
void IsdnDevice::setHangupRequested() {
	pthread_mutex_lock(&mut_hup);
	m_request_hangup = OK;
	pthread_mutex_unlock(&mut_hup);
}

void IsdnDevice::resetHangupRequested() {
	pthread_mutex_lock(&mut_hup);
	m_request_hangup = FAIL;
	pthread_mutex_unlock(&mut_hup);
}

/** Check if call should be ended.
 * @return OK if setClear() was called
 * @return FAIL if setClear() was not called
 */ 
int IsdnDevice::getHangupRequested() {
	int ret = FAIL;
	pthread_mutex_lock(&mut_hup);
	if (m_request_hangup == OK) {
		m_request_hangup = FAIL;
		ret = OK;
	}	
	pthread_mutex_unlock(&mut_hup);
	return ret;
}

/** Open the isdn device file.
 * @return OK on success
 * @return FAIL on any error
 */ 
int IsdnDevice::open_device() {
	m_fd = open(m_devicename, O_RDWR | O_NONBLOCK);
	if (m_fd == -1) {
		ERROR("could not open special file %s, reason: %s\n", m_devicename, strerror(errno));
		return (FAIL);
	}	
	m_status |= STATUS_OPEN;
	return (OK);
}	

/** Setup terminal parameters for isdn device. Special termios settings must
 * be done in order to successfully interact with isdn devices.
 * @return OK on success
 * @return FAIL on any error
 */ 
int IsdnDevice::setTerminalParameters() {
	struct termios tio;
	
	tcgetattr (m_fd, &tio);
	cfsetospeed (&tio, B115200);
	cfsetispeed (&tio, B115200);
	if (tcsetattr (m_fd, TCSANOW, &tio) != 0) {
		ERROR("tcsetattr failed for device %s, reason: %s\n", m_devicename, strerror(errno));
		return FAIL;
	}	
	cfmakeraw (&tio);
	if (tcsetattr (m_fd, TCSANOW, &tio) != 0) {
		ERROR("tcsetattr failed for device %s, reason: %s\n", m_devicename, strerror(errno));
		return FAIL;
	}	
	return OK;
}

/** Removes the lock file for this device.
 * @return OK on success
 * @return FAIL if no lockfile was created or if lockfile could not be removed 
 */ 
int IsdnDevice::removeLockFile() {
	if (m_lockfilename == NULL)
		return (FAIL);
	if ( unlink(m_lockfilename) != 0 ) {
		return (FAIL);
	}
	return (OK);
}
	
/** Create a lock file. The lockfile is placed in the configfile configurable 
 * directory. Stale lockfiles from dead processes are removed with a warning.
 * The lockfile contains the PID of the current process.
 * @return OK on sucess
 * @return FAIL on error
 */ 
int IsdnDevice::createLockFile() {
/* buffer for lockfile issues */
	int lockfd;
	struct stat pidfilestat;
	char buf[BUFSIZE];
	int read_bytes;
	
	if (m_lockfilename == NULL) 
		return (FAIL);
	lockfd = open(m_lockfilename, O_WRONLY | O_CREAT | O_EXCL, 0644);
	if (lockfd < 0 && errno == EEXIST) { 
		pid_t oldpid;
		/* already locked, read file and check if pid is valid */
		lockfd = open(m_lockfilename, O_RDONLY);
		if (lockfd < 0) {
			ERROR("could not read existing pid file %s, reason: %s\n", m_lockfilename, strerror(errno));
			m_lockfilename = NULL;
			return (FAIL);
		}	
		read_bytes = read(lockfd, buf, BUFSIZE);
		if (read_bytes == -1) {
			ERROR("could not read existing pid file %s, reason: %s\n", m_lockfilename, strerror(errno));
			close(lockfd);
			m_lockfilename = NULL;
			return (FAIL);
		}
		buf[read_bytes] = '\0';
		if (sscanf(buf, "%d", &oldpid) != 1) {
			ERROR("failed to read pid from >>%s<<\n", buf);
			close(lockfd);
			m_lockfilename = NULL;
			return (FAIL);
		}
		close(lockfd);
		/* check if process %d is still running */
		snprintf(buf, BUFSIZE, "/proc/%d", oldpid);
		if (stat(buf, &pidfilestat) == 0) {
			ERROR("process %d has an active lock on %s\n", oldpid, m_devicename );
			m_lockfilename = NULL;
			return (FAIL);
		}
		INFO("removing stale lockfile %s\n", m_lockfilename);
		if (unlink(m_lockfilename) != 0) {
			ERROR("could not remove lock file, reason: %s\n", strerror(errno));
			return (FAIL);
		}	
		return (createLockFile());
	} else if (lockfd < 0) {
		ERROR("could not create lockfile %s, reason: %s\n", m_lockfilename, strerror(errno) );
		ERROR("maybe you want to set a different lock directory in configuration file\n");
		m_lockfilename = NULL;
		return (FAIL);
	} else { /* lockfile created */
//		DBG("created lockfile %s\n", m_lockfilename);
		// if (write(lockfd, (char*) getpid(), sizeof(pid_t)) < 0) {	
		snprintf(buf, BUFSIZE, "%d", getpid());
		if (write(lockfd, buf, strlen(buf)) < 0) {	
			ERROR("writing pid to pidfile failed, reason: %s\n", strerror(errno));
			close (lockfd);
			unlink(m_lockfilename);
			return (FAIL);
		}	
		close(lockfd);
		m_status |= STATUS_LOCKED;
		return (OK);
	}
}	

/** Read some bytes from the isdn device. This method is useful only if
 * a voice connection exists. Voice data is read from the device, escape
 * sequences are handled by use of filterSpecials().
 * uses select to poll up to 10 times with a timeout of 2ms.
 * @param buf buffer to be filled
 * @param size number of bytes to read
 * @return number of bytes read on success <BR>
 * @return -1 on read error or end of connection
 */
int IsdnDevice::getVoiceStream(unsigned char *buf, int size) {
	fd_set my_set;
	struct timeval t;
	int count = 0;
	int retries = 0;

	while ( (count < size) && (retries++ < 10) ) { 
		FD_ZERO(&my_set);
		FD_SET(m_fd, &my_set);
		t.tv_sec = 0;
		t.tv_usec = 2000;
		if ( select (m_fd+1, &my_set, NULL, NULL, &t) == 1) {
			/* ready to read */
			int bytes = read (m_fd, buf + count, size - count);
			if (bytes == -1) {
				if ((errno == EAGAIN) && (errno == EINTR)) {
				continue;
				}	
				INFO("read from isdn device failed, reason: %s\n", strerror(errno)); 
				return (-1);
			} else {
				if (filterSpecials(buf+count, &bytes) == FAIL) {
					//DBG("pstn hangup\n");
					*(buf) = 0x10;
					*(buf+1) = 0x14;
					*(buf+2) = 0x10;
					*(buf+3) = 0x03;
					if (putVoiceStream(buf, 4) == FAIL) {
						ERROR("could not end audio connection\n");
						exit(-1);
					}	
					DBG("isdn4linux hangup.\n");
					setHangupRequested();
					m_call->changePSTNState(CALLSTATE_DISCONNECTED);
					return -1;
				}	
			}		
			count += bytes;
		} else {
			continue;
		}	
	}	
	return count;
};

/** Deal with escaped characters. If DTMF signals are of interest, they can be
 * extracted in this function (currently a info message is printed).
 * @param buf buffer to be examined and possibly altered
 * @param size number of valid bytes in buffer
 * @return OK if filtering was successful, buf may be changed <BR>
 * @return FAIL on end of connection
 */
int IsdnDevice::filterSpecials(unsigned char *buf, int *size) {
	unsigned char *result = buf;
	unsigned char *cur = buf;
	unsigned char c;

	while (result - buf != *size) {
		if ( (c = *(cur++)) == 0x10 ) { /* special escape character found */
			/* next char available? TODO: what if buffer ends on 0x10 ??!? */
			if ( (cur - result) != *size) {
				switch (*(cur)) {
					case 0x10 : /* escaped 0x10 -> treat like one 0x10 */
											*(result++) = 0x10;
											cur++;
											break;
					case 0x03 :	/* fall through */
					case 0x14 : /* end of voice connection (requested) */
											DBG("isdn4linux ending voice transfer (0x10 0x14 received)\n");
											return FAIL;
											break;
						default :	/* touch tone */
											if ( (*cur >= '0' && *cur <= '9') || (*cur == '*') || (*cur == '#') ) {
												DBG("touch tone received from pstn >%c< - ignoring.\n", *cur);
											}	
											cur++;
				}
			} else {
				ERROR("FIXME: don't know what to do with a buffer ending on 0x10\n"); /* ignored by now */
			}
		} else { /* normal data byte */
			*(result++) = c;  /* add to output buffer */
		}	
	}
	*size = result - buf;
	return (OK);
}


/** Write some bytes to the isdn device. This method is useful only if
 * a voice connection exists. Voice data is written to the device, escape
 * sequences are not yet handled here. \todo handle escape chars here.
 * @param buf buffer to be written
 * @param size number of bytes to write
 * @return number of bytes written on success <BR>
 * @return -1 on write error
 */

int IsdnDevice::putVoiceStream(unsigned char *buf, int size) {
	int count;
	count = 0;
	while (count < size) { /* more bytes need to be written */
		int bytes = write (m_fd, buf + count, size - count);
		if (bytes == -1) {
			if (errno == EAGAIN) {
				INFO("isdn device not available for writing, dropping data\n");
				return size;
			}	
			INFO("writing voice to isdn device failed, reason: %s\n", strerror(errno)); 
			return (-1);
		}
		count += bytes;
	}	
	
	return count;
};

/** Flushes input and output of the isdn device.
 * This is useful if the state of the device is unknown and AT 
 * commands are be executed.
 */
void IsdnDevice::flush() {
	tcflush(m_fd, TCIFLUSH);
	tcflush(m_fd, TCOFLUSH);
}

