/*
 * $Id: IsdnDeviceOutgoing.cpp,v 1.6 2003/12/04 10:08:26 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "IsdnDevice.h"
#include "IsdnDeviceOutgoing.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "globals.h"
#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"
#include "log.h"
#include "AmSession.h"

/** Create a new IsdnDevice for incoming or outgoing connections. 
 * Sets up synchronization and some other prerequisits, the device is 
 * not opened yet.
 * @param devicename file name of the isdn device, e.g. /dev/ttyI42
 * @param inout type of device (TYPE_OUTGOING or TYPE_INCOMING)
 */
IsdnDeviceOutgoing::IsdnDeviceOutgoing(char* devicename) 
 : IsdnDevice(devicename) 
{
	resetHangupRequested();
	/* initialize the mutex */
	// pthread_mutex_init(&m_mutex, NULL);
}

/** Destructor. Closes the device, removes the lockfile and frees
 * all allocated ressources.
 */
void IsdnDeviceOutgoing::cleanup() {
	//pthread_mutex_unlock(&m_mutex);
	//pthread_mutex_destroy(&m_mutex);
}

void IsdnDeviceOutgoing::onBye() {
	//setHangupRequested();
}	

void IsdnDeviceOutgoing::onBeforeCallAccept(AmRequest* req) {
	char phonenumber[MAX_NUM_LEN];
	char msn[MAX_NUM_LEN];

	resetHangupRequested();
	IsdnUtils::determineNumbers(req, phonenumber, msn, MAX_NUM_LEN); /* throws Exception on error */
	INFO("determine numbers result: from=%s to=%s\n", msn, phonenumber);
	if (strlen(msn) != 0)
		placeCall(req, phonenumber, msn);
	else
		placeCall(req, phonenumber, NULL);
}	

#define resetDevice { hangup(); initialize(); unlock(); resetHangupRequested(); }

void IsdnDeviceOutgoing::placeCall(AmRequest *req, char* phonenumber, char* msn) {
	int ret;
	/* actually place the call */
	if (isdn_config->msnsipcaller && msn != NULL) {
		DBG("using number %s from SIP header for outgoing msn\n", msn);
		ret = dial(phonenumber, msn);
	} else {	
		ret = dial(phonenumber);
	}	

	/* map PSTN beahavior to SIP, error handling, busy ... */
	if (ret == FAIL) {
		resetDevice;
		throw AmSession::Exception(500,"isdngw: sorry, call setup failed");
	} else if (ret == DIALING_STATE_BUSY) {
		resetDevice;
		throw AmSession::Exception(486,"isdngw: busy line");
	} else if (ret == DIALING_STATE_CONNECTED) {
		// SUCCESS!
		return;
	} else if (ret == DIALING_STATE_NODIALTONE) {
		ERROR("received NO DIALTONE from isdn4linux device, recheck your kernel driver configuration\n");
		resetDevice;
		throw AmSession::Exception(500,"isdngw: backend error (isdn4linux: NO DIALTONE)");
	} else if (ret == DIALING_STATE_RINGING) {
		req->reply(183,"isdngw: call progress");
		for (int i=0; i < 200; i++) {
//			char repl[100];
			/* phone is ringing now, wait for an answer */
			/*if (i == 5) *//* send ringing once after some time, no way to determine when the phone really rings :( */
//				req->reply(180,"isdngw: ringing");
			if ((i % 3) == 0)	{
				if (req->reply(180,"isdngw: ringing") == -1) { /* reply failed -> 408! timeout in ser */
					DBG("transaction expired in ser, cancelling PSTN call\n");
					resetDevice;
					return;
				}
			}	
			ret = wait();
			if (getHangupRequested() == OK) {
				DBG("->> hangup requested\n");
				resetDevice;
				throw AmSession::Exception(487,"Call terminated");
				return;
			}		
			if (ret == DIALING_STATE_BUSY) {
				resetDevice;
				throw AmSession::Exception(486,"isdngw: busy line. (1)");
				break;
			} else if (ret == DIALING_STATE_ERROR) {
				resetDevice;
				throw AmSession::Exception(500,"isdngw: sorry, call setup failed.");
				break;
			} else if (ret == DIALING_STATE_CONNECTED) {
				// SUCCESS!
				return;
			} else if (ret == DIALING_STATE_RINGING) {
//				sprintf(repl, "isdngw: ringing (%i/%i)", i, 50);
//				req->reply(180,repl);
//				INFO("I'd really like to send %i %s\n", 180, repl);
			} else	{
				assert(false);
			}	
		}
		resetDevice;
		throw AmSession::Exception(408, "isdngw: request timeout");
	} else {
		assert(false);
	}	
}

void IsdnDeviceOutgoing::onError(unsigned int code, const string& reason)
{
	ERROR("an error occurred - cleaning up isdn device (%i %s)\n", code, reason.c_str());
	hangup();
	initialize();
	unlock();
//	throw AmSession::Exception(500, "isdngw: request timeout");
}




/** Acquire a locked device for outgoing connection.
 * This function tries to find an unused device of the specified type.
 * @return a locked device on success <BR>
 * @return NULL if no device is available
 */
IsdnDevice* IsdnDeviceOutgoing::getLockedDevice() {
	struct list_t *cur = isdn_config->i4l_outdevices;
	int ok = FAIL;
	while (cur != NULL) {
		if ( (cur->num == STATUS_OK) && (((IsdnDevice*)(cur->link))->trylock() == OK) ) {
			ok = OK;
			break;
		} else {
			cur = cur->next;
		}
	}
	if (ok == OK) {
		return (IsdnDevice*)cur->link;
	} else {
		return NULL;
	}	
}

/** Place a call. Try to call the given number with the (optional) specified 
 * msn. This method always returns after no longer than 3 seconds.
 * @return DIALING_STATE_RINGING pstn side is (probably) ringing <BR>
 * @return DIALING_STATE_BUSY if busy signal detected
 * @return DIALING_STATE_CONNECTED if the connection is established
 * @return DIALING_STATE_ERROR if an error occured 
 */ 
int IsdnDeviceOutgoing::dial(char* number, char* msn) {
	//m_clear = FAIL;	
	IsdnChat ic = IsdnChat();
	if (msn == NULL) {
		ic.setOutgoingCall(isdn_config->out_msn, number);
	} else {
		ic.setOutgoingCall(msn, number);
	}
	ic.setTimeout(1, DIALING_STATE_RINGING);
	return (ic.chat(m_fd));
}

/** Wait and see how call proceeds.
 * This method waits up to 2 seconds for a change of the call state.
 * @return DIALING_STATE_RINGING pstn side is (probably) ringing <BR>
 * @return DIALING_STATE_BUSY if busy signal detected
 * @return DIALING_STATE_CONNECTED if the connection is established
 * @return DIALING_STATE_ERROR if an error occured 
 */ 
int IsdnDeviceOutgoing::wait() {
	IsdnChat ic = IsdnChat();
	ic.setWaitForAnswer();
	
	/* return DIALING_STATE_BUSY on BUSY */
	ic.addErrorHandler("BUSY", DIALING_STATE_BUSY);
	/* return DIALING_STATE_CONNECTED if call is answered */
	ic.addErrorHandler("VCON", DIALING_STATE_CONNECTED);
	/* return DIALING_STATE_RINGING if no answer after 1 minute */
	ic.setTimeout(1, DIALING_STATE_RINGING);
	return (ic.chat(m_fd));
}

/** Establish voice connection. State of the device must be IS_CONNECTED.
 */
int IsdnDeviceOutgoing::connectVoice() {
	IsdnChat ic = IsdnChat();
	ic.setConnect();
	return (ic.chat(m_fd));
}

