/*
 * $Id: IsdnGatewayFactory.cpp,v 1.7 2004/04/05 10:07:17 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "globals.h"
#include "IsdnGatewayFactory.h"
#include "IsdnGatewayDialog.h"
#include "IsdnConfiguration.h"
#include "IsdnDevice.h"
#include "IsdnUtils.h"
#include "IsdnDeviceIncoming.h"
#include "IsdnDeviceOutgoing.h"
#include "IsdnChat.h"
#include "IsdnCallManager.h"
#include "IsdnCall.h"
#include "AmAudioIsdn.h"
#include "AmApi.h"
#include "AmSession.h"
#include "AmConfig.h"
#include "AmServer.h"
#include "log.h"


/* register isdngw as a plugin application to sems */
EXPORT_FACTORY(IsdnGatewayFactory,MODULE_NAME);

#ifdef USE_CAPI20
/* global capi manager */
IsdnCapiConnectionManager *capiManager = NULL;
#endif
	

/** isdngw initialization constructor. This constructor is being called once
 * on plugin load. It takes care of 
 *   - checking and loading configuration
 *   - initialize all components
 *   - initialize all isdn devices to be used
 *
 */
IsdnGatewayFactory::IsdnGatewayFactory(const string& _app_name)
  : AmStateFactory(_app_name)
{
	/* load configuration */
	IsdnConfiguration::instantiate();
	isdn_config->reload();

	/* setup call manager */
	IsdnCallManager::instantiate();

	/* deal with isdn4linux ttyI devices */
	if (isdn_config->enablei4l)
		INFO("isdngw: setting up isdn4linux devices\n");
	else
		INFO("isdngw: isdn4linux support disabled in configuration\n");

	/* initilize outgoing devices */
	struct list_t* cur = isdn_config->i4l_outdevices;
	int valid_out_devices = 0;

	while (cur != NULL) {
		cur->link = (void*)new IsdnDeviceOutgoing(cur->item);
		if ( ((IsdnDevice*)(cur->link))->initialize() == FAIL ) {
			ERROR("could not initialize %s, removing from list\n", cur->item);
			delete (IsdnDevice*)cur->link;
			cur->num = STATUS_INVALID;
		} else { 
			cur->num = STATUS_OK;
			valid_out_devices++;
		}	
		cur = cur->next;
	}	

	if ((isdn_config->enablei4l) && (valid_out_devices == 0)) {
		INFO("INFO: no outgoing i4l-devices configured\n");
		isdn_config->i4l_outdevices = NULL;
	}

	/* initilize incoming devices */
	cur = isdn_config->i4l_indevices;
	int valid_in_devices = 0;

	while (cur != NULL) {
		cur->link = (void*) new IsdnDeviceIncoming(cur->item);
		if ( (((IsdnDevice*)cur->link))->initialize() == FAIL ) {
			ERROR("could not initialize %s, removing from list\n", cur->item);
			delete (IsdnDevice*)cur->link;
			cur->num = STATUS_INVALID;
		} else { 
			cur->num = STATUS_OK;
			valid_in_devices++;
		}	
		cur = cur->next;
	}	

	if ((valid_in_devices == 0) && (isdn_config->enablei4l)) {
		INFO("INFO: no incoming i4l-devices configured\n");
		isdn_config->i4l_indevices = NULL;
	}

	IsdnDeviceIncoming::initializeOnce();
	IsdnDeviceIncoming::reassignIncomingDevices();
	
	isdn_config->devices[DEVICE_I4L].numberOfBChannels = valid_in_devices + valid_out_devices;
	if ((isdn_config->enablei4l) && (valid_out_devices == 0)) {
		if ((valid_in_devices == 0) && (isdn_config->enablei4l)) {
			isdn_config->devices[DEVICE_I4L].available = FAIL;
		}
	}	
#ifdef USE_CAPI20
	if (isdn_config->enablecapi) {
		/* deal with CAPI 2.0 devic */
		INFO("INFO: setting up CAPI 2.0 backend\n");

		capiManager = new IsdnCapiConnectionManager();
		capiManager->init();
	}	
#endif
	
/*	isdn_config->debugDumpDeviceList(); */
	isdn_config->dumpConfiguredDevices();
	INFO("isdngw initialization successful.\n");

/* test the matching operations here */
/*
	char* f;
	char* t;
	DBG("running some tests\n");
	f = "81462807";
	t = "81462808";
	DBG("from %s to %s -> device %i\n", f, t, isdn_config->getDeviceForCall(t, f));

	t = "81462808";
	for (int i=0; i<9; i++) 
		DBG("from %s at device %i -- accepted=%s\n", t, i, isdn_config->acceptCallOnDevice(i, t)==OK?"yes":"no");
*/
}



/** A SIP call arrives. This method is called by sems if a SIP INVITE is received,
 * that has to be processed by isdngw.
 */
AmDialogState* IsdnGatewayFactory::onInvite(AmCmd& cmd)
{
	IsdnCall* call;

	/* just determine if incoming or outgoing, then return the dialog */
	
	if (cmd.user.c_str() == NULL)
		throw AmSession::Exception(500,"isdngw: internal error, user is NULL\n");

	if (strncmp(cmd.user.c_str(), INCOMING_KEY_PREFIX, strlen(INCOMING_KEY_PREFIX)) == 0) {
		DBG("received onInvite for incoming call!\n");
		call = callManager->changeSIPState(cmd, CALLSTATE_CONNECTING);
	} else {
		DBG("received onInvite for outgoing call!\n");
		call = callManager->placeCallFromSIP(cmd);
	}
	//DBG("now returning GatewayDialog\n");
	return new IsdnGatewayDialog(call);

}

IsdnGatewayFactory::~IsdnGatewayFactory()
{
	DBG("isdngw: destructor of IsdnGatewayFactory: cleaning up.\n");
	delete(isdn_config);
	delete(callManager);
}

int IsdnGatewayFactory::onLoad() {
	DBG("registering fifo function %s for web access...\n", FIFO_WEB_FUNC);
	AmServer::instance()->registerFct(FIFO_WEB_FUNC, callManager);
	return 0;
}
